"""Hadamard matrix."""
import numpy as np


def hadamard(n_param: int = 1) -> np.ndarray:
    r"""
    Produce a 2^{n_param} dimensional Hadamard matrix [WikHad]_.

    The standard Hadamard matrix that is often used in quantum information as a
    two-qubit quantum gate is defined as

    .. math::
        H_1 = \frac{1}{\sqrt{2}} \begin{pmatrix}
                                    1 & 1 \\
                                    1 & -1
                                 \end{pmatrix}

    In general, the Hadamard matrix of dimension :math:2^{n_param} may be
    defined as

    .. math::
        \left( H_n \right)_{i, j} = \frac{1}{2^{\frac{n}{2}}}
        \left(-1\right)^{i \dot j}

    Examples
    ==========

    The standard 2-qubit Hadamard matrix can be generated in `toqito` as

    >>> from toqito.matrices import hadamard
    >>> hadamard(1)
    [[ 0.70710678  0.70710678]
     [ 0.70710678 -0.70710678]]

    References
    ==========
    .. [WikHad] Wikipedia: Hadamard transform
        https://en.wikipedia.org/wiki/Hadamard_transform

    :param n_param: A non-negative integer (default = 1).
    :return: The Hadamard matrix of dimension `2^{n_param}`.
    """
    if n_param == 0:
        return np.array([1])
    if n_param == 1:
        return 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]])
    if n_param == 2:
        return 1 / 2 * np.array([[1, 1, 1, 1], [1, -1, 1, -1], [1, 1, -1, -1], [1, -1, -1, 1]])
    if n_param > 1:
        mat_1 = hadamard(1)
        mat_2 = hadamard(1)
        mat = np.kron(mat_1, mat_2)
        for _ in range(2, n_param):
            mat_1 = mat_2
            mat_2 = mat
            mat = np.kron(mat_1, mat_2)
        return mat
    raise ValueError(f"Improper dimension {n_param} provided.")



"""Permutation operator."""
from typing import List, Union

import numpy as np

from toqito.matrices import iden
from toqito.perms import permute_systems


def permutation_operator(
    dim: Union[List[int], int],
    perm: List[int],
    inv_perm: bool = False,
    is_sparse: bool = False,
) -> np.ndarray:
    r"""
    Produce a unitary operator that permutes subsystems.

    Generates a unitary operator that permutes the order of subsystems according to the permutation
    vector :code:`perm`, where the :math:`i^{th}` subsystem has dimension :code:`dim[i]`.

    If :code:`inv_perm` = True, it implements the inverse permutation of :code:`perm`. The
    permutation operator return is full is :code:`is_sparse` is :code:`False` and sparse if
    :code:`is_sparse` is :code:`True`.

    Examples
    ==========

    The permutation operator obtained with dimension :math:`d = 2` is equivalent to the standard
    swap operator on two qubits

    .. math::
        P_{2, [2, 1]} =
        \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 0 & 1 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 0 & 1
        \end{pmatrix}

    Using :code:`toqito`, this can be achieved in the following manner.

    >>> from toqito.perms import permutation_operator
    >>> permutation_operator(2, [2, 1])
    [[1., 0., 0., 0.],
     [0., 0., 1., 0.],
     [0., 1., 0., 0.],
     [0., 0., 0., 1.]]

    :param dim: The dimensions of the subsystems to be permuted.
    :param perm: A permutation vector.
    :param inv_perm: Boolean dictating if :code:`perm` is inverse or not.
    :param is_sparse: Boolean indicating if return is sparse or not.
    :return: Permutation operator of dimension :code:`dim`.
    """
    # Allow the user to enter a single number for `dim`.
    if isinstance(dim, int):
        dim = dim * np.ones(max(perm))
    if isinstance(dim, list):
        dim = np.array(dim)

    # Swap the rows of the identity matrix appropriately.
    return permute_systems(iden(int(np.prod(dim)), is_sparse), perm, dim, True, inv_perm)



"""A number of operations on quantum channels."""
from toqito.channel_ops.apply_channel import apply_channel
from toqito.channel_ops.partial_channel import partial_channel
from toqito.channel_ops.choi_to_kraus import choi_to_kraus
from toqito.channel_ops.kraus_to_choi import kraus_to_choi



"""Test swap."""
import numpy as np

from toqito.perms import swap


def test_swap_matrix():
    """Tests swap operation on matrix."""
    test_mat = np.array([[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]])

    expected_res = np.array([[1, 9, 5, 13], [3, 11, 7, 15], [2, 10, 6, 14], [4, 12, 8, 16]])

    res = swap(test_mat)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_swap_vector_1():
    """Tests swap operation on vector."""
    test_vec = np.array([1, 2, 3, 4])

    expected_res = np.array([1, 3, 2, 4])

    res = swap(test_vec)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_swap_int_dim():
    """Test swap operation when int is provided."""
    test_mat = np.array([[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]])

    expected_res = np.array([[1, 9, 5, 13], [3, 11, 7, 15], [2, 10, 6, 14], [4, 12, 8, 16]])

    res = swap(test_mat, [1, 2], 2)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_swap_8_by_8():
    """Test swap on an 8x8 matrix."""
    test_mat = np.arange(64).reshape(8, 8).T

    expected_res = np.array(
        [
            [0, 16, 8, 24, 32, 48, 40, 56],
            [2, 18, 10, 26, 34, 50, 42, 58],
            [1, 17, 9, 25, 33, 49, 41, 57],
            [3, 19, 11, 27, 35, 51, 43, 59],
            [4, 20, 12, 28, 36, 52, 44, 60],
            [6, 22, 14, 30, 38, 54, 46, 62],
            [5, 21, 13, 29, 37, 53, 45, 61],
            [7, 23, 15, 31, 39, 55, 47, 63],
        ]
    )
    res = swap(test_mat, [2, 3], [2, 2, 2])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_swap_invalid_dim():
    """Invalid dim parameters."""
    test_mat = np.array([[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]])
    with np.testing.assert_raises(ValueError):
        swap(test_mat, [1, 2], 5)


def test_swap_invalid_sys():
    """Invalid sys parameters."""
    test_mat = np.array([[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]])
    with np.testing.assert_raises(ValueError):
        swap(test_mat, [0])


def test_swap_invalid_sys_len():
    """Invalid sys parameters."""
    test_mat = np.array([[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]])
    with np.testing.assert_raises(ValueError):
        swap(test_mat, [1], 2)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Von neumann entropy metric."""
import numpy as np

from toqito.matrix_props import is_density


def von_neumann_entropy(rho: np.ndarray) -> float:
    r"""
    Compute the von Neumann entropy of a density matrix [WikVent]_. [WatVec]_.

    Let :math:`P \in \text{Pos}(\mathcal{X})` be a positive semidefinite operator, for a complex
    Euclidean space :math:`\mathcal{X}`. Then one defines the *von Neumann entropy* as

    .. math::
        H(P) = H(\lambda(P)),

    where :math:`\lambda(P)` is the vector of eigenvalues of :math:`P` and where the function
    :math:`H(\cdot)` is the Shannon entropy function defined as

    .. math::
        H(u) = -\sum_{\substack{a \in \Sigma \\ u(a) > 0}} u(a) \text{log}(u(a)),

    where the :math:`\text{log}` function is assumed to be the base-2 logarithm, and where
    :math:`\Sigma` is an alphabet where :math:`u \in [0, \infty)^{\Sigma}` is a vector of
    nonnegative real numbers indexed by :math:`\Sigma`.

    Examples
    ==========

    Consider the following Bell state:

    .. math::
        u = \frac{1}{\sqrt{2}} \left(|00 \rangle + |11 \rangle \right) \in \mathcal{X}.

    The corresponding density matrix of :math:`u` may be calculated by:

    .. math::
        \rho = u u^* = \frac{1}{2} \begin{pmatrix}
                         1 & 0 & 0 & 1 \\
                         0 & 0 & 0 & 0 \\
                         0 & 0 & 0 & 0 \\
                         1 & 0 & 0 & 1
                       \end{pmatrix} \text{D}(\mathcal{X}).

    Calculating the von Neumann entropy of :math:`\rho` in :code:`toqito` can be done as follows.

    >>> from toqito.state_props import von_neumann_entropy
    >>> import numpy as np
    >>> test_input_mat = np.array(
    >>>     [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0],
    >>>      [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
    >>> )
    >>> von_neumann_entropy(test_input_mat)
    5.88418203051333e-15

    Consider the density operator corresponding to the maximally mixed state of dimension two

    .. math::
        \rho = \frac{1}{2}
        \begin{pmatrix}
            1 & 0 \\
            0 & 1
        \end{pmatrix}.

    As this state is maximally mixed, the von Neumann entropy of :math:`\rho` is
    equal to one. We can see this in :code:`toqito` as follows.

    >>> from toqito.state_props import von_neumann_entropy
    >>> import numpy as np
    >>> rho = 1/2 * np.identity(2)
    >>> von_neumann_entropy(rho)
    1.0

    References
    ==========
    .. [WikVent] Wikipedia: Von Neumann entropy
        https://en.wikipedia.org/wiki/Von_Neumann_entropy

    .. [WatVec] Watrous, John.
        "The theory of quantum information."
        Section: "Definitions of quantum entropic functions".
        Cambridge University Press, 2018.

    :param rho: Density operator.
    :return: The von Neumann entropy of :code:`rho`.
    """
    if not is_density(rho):
        raise ValueError("Von Neumann entropy is only defined for density " "operators.")
    eigs, _ = np.linalg.eig(rho)
    eigs = [eig for eig in eigs if eig > 0]
    return -np.sum(np.real(eigs * np.log2(eigs)))



"""The partial transpose."""
from typing import List, Union

import numpy as np

from cvxpy.expressions.expression import Expression
from cvxpy.expressions.variable import Variable
from toqito.perms import permute_systems
from toqito.helper import expr_as_np_array, np_array_as_expr


def partial_transpose(
    rho: Union[np.ndarray, Variable],
    sys: Union[List[int], np.ndarray, int] = 2,
    dim: Union[List[int], np.ndarray] = None,
) -> Union[np.ndarray, Expression]:
    r"""Compute the partial transpose of a matrix [WikPtrans]_.

    The *partial transpose* is defined as

    .. math::
        \left( \text{T} \otimes \mathbb{I}_{\mathcal{Y}} \right)
        \left(X\right)

    where :math:`X \in \text{L}(\mathcal{X})` is a linear operator over the complex Euclidean
    space :math:`\mathcal{X}` and where :math:`\text{T}` is the transpose mapping
    :math:`\text{T} \in \text{T}(\mathcal{X})` defined as

    .. math::
        \text{T}(X) = X^{\text{T}}

    for all :math:`X \in \text{L}(\mathcal{X})`.

    By default, the returned matrix is the partial transpose of the matrix :code:`rho`, where it
    is assumed that the number of rows and columns of :code:`rho` are both perfect squares and
    both subsystems have equal dimension. The transpose is applied to the second subsystem.

    In the case where :code:`sys` amd :code:`dim` are specified, this function gives the partial
    transpose of the matrix :code:`rho` where the dimensions of the (possibly more than 2)
    subsystems are given by the vector :code:`dim` and the subsystems to take the partial
    transpose are given by the scalar or vector :code:`sys`. If :code:`rho` is non-square,
    different row and column dimensions can be specified by putting the row dimensions in the
    first row of :code:`dim` and the column dimensions in the second row of :code:`dim`.

    Examples
    ==========

    Consider the following matrix

    .. math::
        X = \begin{pmatrix}
                1 & 2 & 3 & 4 \\
                5 & 6 & 7 & 8 \\
                9 & 10 & 11 & 12 \\
                13 & 14 & 15 & 16
            \end{pmatrix}.

    Performing the partial transpose on the matrix :math:`X` over the second
    subsystem yields the following matrix

    .. math::
        X_{pt, 2} = \begin{pmatrix}
                    1 & 5 & 3 & 7 \\
                    2 & 6 & 4 & 8 \\
                    9 & 13 & 11 & 15 \\
                    10 & 14 & 12 & 16
                 \end{pmatrix}.

    By default, in :code:`toqito`, the partial transpose function performs the transposition on
    the second subsystem as follows.

    >>> from toqito.channels import partial_transpose
    >>> import numpy as np
    >>> test_input_mat = np.arange(1, 17).reshape(4, 4)
    >>> partial_transpose(test_input_mat)
    [[ 1  5  3  7]
     [ 2  6  4  8]
     [ 9 13 11 15]
     [10 14 12 16]]

    By specifying the :code:`sys = 1` argument, we can perform the partial transpose over the
    first subsystem (instead of the default second subsystem as done above). Performing the
    partial transpose over the first subsystem yields the following matrix

    .. math::
        X_{pt, 1} = \begin{pmatrix}
                        1 & 2 & 9 & 10 \\
                        5 & 6 & 13 & 14 \\
                        3 & 4 & 11 & 12 \\
                        7 & 8 & 15 & 16
                    \end{pmatrix}.

    >>> from toqito.channels import partial_transpose
    >>> import numpy as np
    >>> test_input_mat = np.array(
    >>>     [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
    >>> )
    >>> partial_transpose(test_input_mat, 1)
    [[ 1  2  9 10]
     [ 5  6 13 14]
     [ 3  4 11 12]
     [ 7  8 15 16]]

    References
    ==========
    .. [WikPtrans] Wikipedia: Partial transpose
        https://en.wikipedia.org/w/index.php?title=Partial_transpose

    :param rho: A matrix.
    :param sys: Scalar or vector specifying the size of the subsystems.
    :param dim: Dimension of the subsystems. If :code:`None`, all dimensions
                are assumed to be equal.
    :returns: The partial transpose of matrix :code:`rho`.
    """
    # If the input matrix is a CVX variable for an SDP, we convert it to a
    # numpy array, perform the partial transpose, and convert it back to a CVX
    # variable.
    if isinstance(rho, Variable):
        rho_np = expr_as_np_array(rho)
        transposed_rho = partial_transpose(rho_np, sys, dim)
        transposed_rho = np_array_as_expr(transposed_rho)
        return transposed_rho

    sqrt_rho_dims = np.round(np.sqrt(list(rho.shape)))

    if dim is None:
        dim = np.array([[sqrt_rho_dims[0], sqrt_rho_dims[0]], [sqrt_rho_dims[1], sqrt_rho_dims[1]]])
    if isinstance(dim, float):
        dim = np.array([dim])
    if isinstance(dim, list):
        dim = np.array(dim)
    if isinstance(sys, list):
        sys = np.array(sys)
    if isinstance(sys, int):
        sys = np.array([sys])

    num_sys = len(dim)
    # Allow the user to enter a single number for dim.
    if num_sys == 1:
        dim = np.array([dim, list(rho.shape)[0] / dim])
        if np.abs(dim[1] - np.round(dim[1]))[0] >= 2 * list(rho.shape)[0] * np.finfo(float).eps:
            raise ValueError(
                "InvalidDim: If `dim` is a scalar, `rho` must be "
                "square and `dim` must evenly divide `len(rho)`; "
                "please provide the `dim` array containing the "
                "dimensions of the subsystems."
            )
        dim[1] = np.round(dim[1])
        num_sys = 2

    # Allow the user to enter a vector for dim if X is square.
    if min(dim.shape) == 1 or len(dim.shape) == 1:
        # Force dim to be a row vector.
        dim = dim.T.flatten()
        dim = np.array([dim, dim])

    prod_dim_r = int(np.prod(dim[0, :]))
    prod_dim_c = int(np.prod(dim[1, :]))

    sub_prod_r = np.prod(dim[0, sys - 1])
    sub_prod_c = np.prod(dim[1, sys - 1])

    sub_sys_vec_r = prod_dim_r * np.ones(int(sub_prod_r)) / sub_prod_r
    sub_sys_vec_c = prod_dim_c * np.ones(int(sub_prod_c)) / sub_prod_c

    set_diff = list(set(list(range(1, num_sys + 1))) - set(sys))
    perm = sys.tolist()[:]
    perm.extend(set_diff)

    # Permute the subsystems so that we just have to do the partial transpose
    # on the first (potentially larger) subsystem.
    rho_permuted = permute_systems(rho, perm, dim)

    x_tmp = np.reshape(
        rho_permuted,
        [
            int(sub_sys_vec_r[0]),
            int(sub_prod_r),
            int(sub_sys_vec_c[0]),
            int(sub_prod_c),
        ],
        order="F",
    )
    y_tmp = np.transpose(x_tmp, [0, 3, 2, 1])
    z_tmp = np.reshape(
        y_tmp,
        [
            int(sub_sys_vec_r[0]) * int(sub_prod_c),
            int(sub_sys_vec_c[0]) * int(sub_prod_r),
        ],
        order="F",
    )

    # If z_tmp is a just a 1-D matrix, extract just the row.
    if z_tmp.shape[0] == 1:
        z_tmp = z_tmp[0]

    # Return the subsystems back to their original positions.
    dim[[0, 1], sys - 1] = dim[[1, 0], sys - 1]

    return permute_systems(z_tmp, perm, dim, False, True)



"""Tests for is_unital."""
import numpy as np

from toqito.channel_props import is_unital
from toqito.channels import depolarizing
from toqito.perms import swap_operator


def test_is_unital_kraus_true():
    """Verify unital channel as Kraus ops as True."""
    kraus_1 = np.array([[1, 0], [0, 0]])
    kraus_2 = np.array([[1, 0], [0, 0]]).conj().T
    kraus_3 = np.array([[0, 1], [0, 0]])
    kraus_4 = np.array([[0, 1], [0, 0]]).conj().T
    kraus_5 = np.array([[0, 0], [1, 0]])
    kraus_6 = np.array([[0, 0], [1, 0]]).conj().T
    kraus_7 = np.array([[0, 0], [0, 1]])
    kraus_8 = np.array([[0, 0], [0, 1]]).conj().T

    kraus_ops = [
        [kraus_1, kraus_2],
        [kraus_3, kraus_4],
        [kraus_5, kraus_6],
        [kraus_7, kraus_8],
    ]
    np.testing.assert_equal(is_unital(kraus_ops), True)


def test_is_unital_swap_operator_choi_true():
    """Verify Choi matrix of the swap operator map is unital."""
    np.testing.assert_equal(is_unital(swap_operator(3)), True)


def test_is_unital_depolarizing_choi_false():
    """Verify Choi matrix of the depolarizing map is not unital."""
    np.testing.assert_equal(is_unital(depolarizing(4)), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Tests for sub_fidelity."""
import numpy as np


from toqito.state_metrics import fidelity
from toqito.state_metrics import sub_fidelity
from toqito.states import basis
from toqito.states import bell


def test_sub_fidelity_default():
    """Test sub_fidelity default arguments."""
    rho = bell(0) * bell(0).conj().T
    sigma = rho

    res = sub_fidelity(rho, sigma)
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_sub_fidelity_lower_bound_1():
    """Test sub_fidelity is lower bound on fidelity for rho and sigma."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    rho = 3 / 4 * e_0 * e_0.conj().T + 1 / 4 * e_1 * e_1.conj().T
    sigma = 2 / 3 * e_0 * e_0.conj().T + 1 / 3 * e_1 * e_1.conj().T

    res = sub_fidelity(rho, sigma)
    np.testing.assert_array_less(res, fidelity(rho, sigma))


def test_sub_fidelity_lower_bound_2():
    """Test sub_fidelity is lower bound on fidelity for rho and pi."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    rho = 3 / 4 * e_0 * e_0.conj().T + 1 / 4 * e_1 * e_1.conj().T
    sigma = 1 / 8 * e_0 * e_0.conj().T + 7 / 8 * e_1 * e_1.conj().T

    res = sub_fidelity(rho, sigma)
    np.testing.assert_array_less(res, fidelity(rho, sigma))


def test_non_square_sub_fidelity():
    """Tests for invalid dim for sub_fidelity."""
    rho = np.array([[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [1, 0, 0, 1]])
    sigma = np.array([[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [2, 0, 0, 2]])
    with np.testing.assert_raises(ValueError):
        sub_fidelity(rho, sigma)


def test_sub_fidelity_invalid_dim_sub_fidelity():
    """Tests for invalid dim for sub_fidelity."""
    rho = np.array([[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]])
    sigma = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    with np.testing.assert_raises(ValueError):
        sub_fidelity(rho, sigma)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test vec."""
import numpy as np

from toqito.matrix_ops import vec


def test_vec():
    """Test standard vec operation on a matrix."""
    expected_res = np.array([[1], [3], [2], [4]])

    test_input_mat = np.array([[1, 2], [3, 4]])

    res = vec(test_input_mat)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Maximally mixed state."""
import numpy as np

from scipy import sparse


def max_mixed(dim: int, is_sparse: bool = False) -> [np.ndarray, sparse.dia.dia_matrix]:
    r"""
    Produce the maximally mixed state [AAR6]_.

    Produces the maximally mixed state on of :code:`dim` dimensions. The maximally mixed state is
    defined as

    .. math::
        \omega = \frac{1}{d} \begin{pmatrix}
                        1 & 0 & \ldots & 0 \\
                        0 & 1 & \ldots & 0 \\
                        \vdots & \vdots & \ddots & \vdots \\
                        0 & 0 & \ldots & 1
                    \end{pmatrix},

    or equivalently, it is defined as

    .. math::
        \omega = \frac{\mathbb{I}}{\text{dim}(\mathcal{X})}

    for some complex Euclidean space :math:`\mathcal{X}`. The maximally mixed state is sometimes
    also referred to as the tracial state.

    The maximally mixed state is returned as a sparse matrix if :code:`is_sparse = True` and is full
    if :code:`is_sparse = False`.

    Examples
    ==========

    Using :code:`toqito`, we can generate the :math:`2`-dimensional maximally mixed state

    .. math::
        \omega_2 = \frac{1}{2}
        \begin{pmatrix}
            1 & 0 \\
            0 & 1
        \end{pmatrix}

    as follows.

    >>> from toqito.states import max_mixed
    >>> max_mixed(2, is_sparse=False)
    [[0.5, 0. ],
     [0. , 0.5]]

    One may also generate a maximally mixed state returned as a sparse matrix

    >>> from toqito.states import max_mixed
    >>> max_mixed(2, is_sparse=True)
        <2x2 sparse matrix of type '<class 'numpy.float64'>'
        with 2 stored elements (1 diagonals) in DIAgonal format>

    References
    ==========
    .. [AAR6] Scott Aaronson: Lecture 6, Thurs Feb 2: Mixed States
        https://www.scottaaronson.com/qclec/6.pdf

    :param dim: Dimension of the entangled state.
    :param is_sparse: `True` if vector is spare and `False` otherwise.
    :return: The maximally mixed state of dimension `dim`.
    """
    if is_sparse:
        return 1 / dim * sparse.eye(dim)
    return 1 / dim * np.eye(dim)



"""Test reduction."""
import numpy as np

from toqito.channels import reduction


def test_reduction_map():
    """Test for the standard reduction map."""
    res = reduction(3)
    np.testing.assert_equal(res[4, 0], -1)
    np.testing.assert_equal(res[8, 0], -1)
    np.testing.assert_equal(res[1, 1], 1)
    np.testing.assert_equal(res[2, 2], 1)
    np.testing.assert_equal(res[3, 3], 1)
    np.testing.assert_equal(res[0, 4], -1)
    np.testing.assert_equal(res[8, 4], -1)
    np.testing.assert_equal(res[5, 5], 1)
    np.testing.assert_equal(res[6, 6], 1)
    np.testing.assert_equal(res[7, 7], 1)
    np.testing.assert_equal(res[0, 8], -1)
    np.testing.assert_equal(res[4, 8], -1)


def test_reduction_map_dim_3_k_2():
    """Test for the reduction map with dimension 3 and parameter k = 2."""
    res = reduction(3, 2)
    expected_res = np.array(
        [
            [1, 0, 0, 0, -1, 0, 0, 0, -1],
            [0, 2, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 2, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 2, 0, 0, 0, 0, 0],
            [-1, 0, 0, 0, 1, 0, 0, 0, -1],
            [0, 0, 0, 0, 0, 2, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 2, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 2, 0],
            [-1, 0, 0, 0, -1, 0, 0, 0, 1],
        ]
    )

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Generate random density matrix."""
from typing import List, Union

import numpy as np

from toqito.random import random_unitary


def random_density_matrix(
    dim: int,
    is_real: bool = False,
    k_param: Union[List[int], int] = None,
    distance_metric: str = "haar",
) -> np.ndarray:
    r"""
    Generate a random density matrix.

    Generates a random :code:`dim`-by-:code:`dim` density matrix distributed according to the
    Hilbert-Schmidt measure. The matrix is of rank <= :code:`k_param` distributed according to the
    distribution :code:`distance_metric` If :code:`is_real = True`, then all of its entries will be
    real. The variable :code:`distance_metric` must be one of:

        - :code:`haar` (default):
            Generate a larger pure state according to the Haar measure and
            trace out the extra dimensions. Sometimes called the
            Hilbert-Schmidt measure when :code:`k_param = dim`.

        - :code:`bures`:
            The Bures measure.

    Examples
    ==========

    Using :code:`toqito`, we may generate a random complex-valued :math:`n`- dimensional density
    matrix. For :math:`d=2`, this can be accomplished as follows.

    >>> from toqito.random import random_density_matrix
    >>> complex_dm = random_density_matrix(2)
    >>> complex_dm
    [[0.34903796+0.j       0.4324904 +0.103298j]
     [0.4324904 -0.103298j 0.65096204+0.j      ]]

    We can verify that this is in fact a valid density matrix using the :code:`is_denisty` function
    from :code:`toqito` as follows

    >>> from toqito.matrix_props import is_density
    >>> is_density(complex_dm)
    True

    We can also generate random density matrices that are real-valued as follows.

    >>> from toqito.random import random_density_matrix
    >>> real_dm = random_density_matrix(2, is_real=True)
    >>> real_dm
    [[0.37330805 0.46466224]
     [0.46466224 0.62669195]]

    Again, verifying that this is a valid density matrix can be done as follows.

    >>> from toqito.matrix_props import is_density
    >>> is_density(real_dm)
    True

    By default, the random density operators are constructed using the Haar measure. We can select
    to generate the random density matrix according to the Bures metric instead as follows.

    >>> from toqito.random import random_density_matrix
    >>> bures_mat = random_density_matrix(2, distance_metric="bures")
    >>> bures_mat
    [[0.59937164+0.j         0.45355087-0.18473365j]
     [0.45355087+0.18473365j 0.40062836+0.j        ]]

    As before, we can verify that this matrix generated is a valid density matrix.

    >>> from toqito.matrix_props import is_density
    >>> is_density(bures_mat)
    True

    :param dim: The number of rows (and columns) of the density matrix.
    :param is_real: Boolean denoting whether the returned matrix will have all
                    real entries or not.
    :param k_param: Default value is equal to :code:`dim`.
    :param distance_metric: The distance metric used to randomly generate the
                            density matrix. This metric is either the Haar
                            measure or the Bures measure. Default value is to
                            use the Haar measure.
    :return: A :code:`dim`-by-:code:`dim` random density matrix.
    """
    if k_param is None:
        k_param = dim

    # Haar / Hilbert-Schmidt measure.
    gin = np.random.rand(dim, k_param)

    if not is_real:
        gin = gin + 1j * np.random.rand(dim, k_param)

    if distance_metric == "bures":
        gin = random_unitary(dim, is_real) + np.identity(dim) @ gin

    rho = gin @ np.array(gin).conj().T

    return np.divide(rho, np.trace(rho))



"""Tests for choi_rank."""
import numpy as np
import pytest

from toqito.channel_props import choi_rank


def test_choi_rank_list_kraus():
    """Verify that a list of Kraus operators gives correct Choi rank"""
    kraus_1 = np.array([[1, 0], [0, 0]])
    kraus_2 = np.array([[1, 0], [0, 0]]).conj().T
    kraus_3 = np.array([[0, 1], [0, 0]])
    kraus_4 = np.array([[0, 1], [0, 0]]).conj().T
    kraus_5 = np.array([[0, 0], [1, 0]])
    kraus_6 = np.array([[0, 0], [1, 0]]).conj().T
    kraus_7 = np.array([[0, 0], [0, 1]])
    kraus_8 = np.array([[0, 0], [0, 1]]).conj().T

    kraus_ops = [
        [kraus_1, kraus_2],
        [kraus_3, kraus_4],
        [kraus_5, kraus_6],
        [kraus_7, kraus_8],
    ]
    np.testing.assert_equal(choi_rank(kraus_ops), 4)


def test_choi_rank_choi_matrix():
    """Verify Choi matrix of the swap operator map gives correct Choi rank."""
    choi_matrix = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])
    np.testing.assert_equal(choi_rank(choi_matrix), 4)


def test_choi_bad_input():
    """Verify that a bad input (such as a string which still passes
    with `numpy.linalg.matrix_rank`) raises an error"""
    with pytest.raises(ValueError, match="Not a valid"):
        bad_input = "string"
        choi_rank(bad_input)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""The realignment channel."""
import numpy as np

from toqito.perms import swap
from toqito.channels import partial_transpose


def realignment(input_mat: np.ndarray, dim=None) -> np.ndarray:
    r"""
    Compute the realignment of a bipartite operator [LAS08]_.

    Gives the realignment of the matrix :code:`input_mat`, where it is assumed that the number
    of rows and columns of :code:`input_mat` are both perfect squares and both subsystems have
    equal dimension. The realignment is defined by mapping the operator :math:`|ij \rangle
    \langle kl |` to :math:`|ik \rangle \langle jl |` and extending linearly.

    If :code:`input_mat` is non-square, different row and column dimensions can be specified by
    putting the row dimensions in the first row of :code:`dim` and the column dimensions in the
    second row of :code:`dim`.

    Examples
    ==========

    The standard realignment map

    Using :code:`toqito`, we can generate the standard realignment map as follows. When viewed as a
    map on block matrices, the realignment map takes each block of the original matrix and makes
    its vectorization the rows of the realignment matrix. This is illustrated by the following
    small example:

    >>> from toqito.channels import realignment
    >>> import numpy as np
    >>> test_input_mat = np.array(
    >>>     [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
    >>> )
    >>> realignment(test_input_mat)
    [[ 1  2  5  6]
     [ 3  4  7  8]
     [ 9 10 13 14]
     [11 12 15 16]]

    References
    ==========
    .. [LAS08] Lupo, Cosmo, Paolo, Aniello, and Scardicchio, Antonello.
        "Bipartite quantum systems: on the realignment criterion and beyond."
        Journal of Physics A: Mathematical and Theoretical
        41.41 (2008): 415301.
        https://arxiv.org/abs/0802.2019

    :param input_mat: The input matrix.
    :param dim: Default has all equal dimensions.
    :return: The realignment map matrix.
    """
    eps = np.finfo(float).eps
    dim_mat = input_mat.shape
    round_dim = np.round(np.sqrt(dim_mat))
    if dim is None:
        dim = np.transpose(np.array([round_dim]))
    if isinstance(dim, list):
        dim = np.array(dim)

    if isinstance(dim, int):
        dim = np.array([int(dim), int(dim_mat[0] / dim)])
        if np.abs(dim[1] - np.round(dim[1])) >= 2 * dim_mat[0] * eps:
            raise ValueError("InvalidDim:")
        dim[1] = np.round(dim[1])

    # Dimension if row vector.
    if len(dim.shape) == 1:
        dim = dim[:].T
        dim = np.array([dim, dim])

    # Dimension is column vector.
    if min(dim.shape) == 1:
        dim = dim[:].T[0]
        dim = np.array([dim, dim])

    dim_x = np.array([[dim[0][1], dim[0][0]], [dim[1][0], dim[1][1]]])
    dim_y = np.array([[dim[1][0], dim[0][0]], [dim[0][1], dim[1][1]]])

    x_tmp = swap(input_mat, [1, 2], dim, True)
    y_tmp = partial_transpose(x_tmp, sys=1, dim=dim_x)

    return swap(y_tmp, [1, 2], dim_y, True)



"""Test is_hermitian."""
import numpy as np

from toqito.matrix_props import is_hermitian


def test_is_hermitian():
    """Test if matrix is Hermitian."""
    mat = np.array([[2, 2 + 1j, 4], [2 - 1j, 3, 1j], [4, -1j, 1]])
    np.testing.assert_equal(is_hermitian(mat), True)


def test_is_non_hermitian():
    """Test non-Hermitian matrix."""
    mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    np.testing.assert_equal(is_hermitian(mat), False)


def test_is_hermitian_not_square():
    """Input must be a square matrix."""
    mat = np.array([[-1, 1, 1], [1, 2, 3]])
    np.testing.assert_equal(is_hermitian(mat), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Is channel positive."""
from typing import List, Union

import numpy as np

from toqito.channel_ops import kraus_to_choi
from toqito.matrix_props import is_positive_semidefinite


def is_positive(
    phi: Union[np.ndarray, List[List[np.ndarray]]],
    rtol: float = 1e-05,
    atol: float = 1e-08,
) -> bool:
    r"""
    Determine whether the given channel is positive [WatPM18]_.

    A map :math:`\Phi \in \text{T} \left(\mathcal{X}, \mathcal{Y} \right)` is *positive* if it
    holds that

    .. math::
        \Phi(P) \in \text{Pos}(\mathcal{Y})

    for every positive semidefinite operator :math:`P \in \text{Pos}(\mathcal{X})`.

    Alternatively, a channel is positive if the corresponding Choi matrix of the channel is both
    Hermitian-preserving and positive semidefinite.

    Examples
    ==========

    We can specify the input as a list of Kraus operators. Consider the map :math:`\Phi` defined as

    .. math::
        \Phi(X) = X - U X U^*

    where

    .. math::
        U = \frac{1}{\sqrt{2}}
        \begin{pmatrix}
            1 & 1 \\
            -1 & -1
        \end{pmatrix}.

    This map is not completely positive, as we can verify as follows.

    >>> from toqito.channel_props import is_positive
    >>> import numpy as np
    >>> unitary_mat = np.array([[1, 1], [-1, -1]]) / np.sqrt(2)
    >>> kraus_ops = [[np.identity(2), np.identity(2)], [unitary_mat, -unitary_mat]]
    >>> is_positive(kraus_ops)
    False

    We can also specify the input as a Choi matrix. For instance, consider the Choi matrix
    corresponding to the :math:`4`-dimensional completely depolarizing channel and may verify
    that this channel is positive.

    >>> from toqito.channels import depolarizing
    >>> from toqito.channel_props import is_positive
    >>> is_positive(depolarizing(4))
    True

    References
    ==========
    .. [WatPM18] Watrous, John.
        "The theory of quantum information."
        Section: "Linear maps of square operators".
        Cambridge University Press, 2018.

    :param phi: The channel provided as either a Choi matrix or a list of Kraus operators.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: True if the channel is positive, and False otherwise.
    """
    # If the variable `phi` is provided as a list, we assume this is a list
    # of Kraus operators.
    if isinstance(phi, list):
        phi = kraus_to_choi(phi)
    return is_positive_semidefinite(phi, rtol, atol)



"""Test l1_norm_coherence."""
import numpy as np

from toqito.state_props import l1_norm_coherence


def test_l1_norm_coherence_maximally_coherence():
    """The l1-norm coherence of the maximally coherent state."""
    v_vec = np.ones((3, 1)) / np.sqrt(3)
    res = l1_norm_coherence(v_vec)

    np.testing.assert_equal(np.isclose(res, 2), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test brauer."""
import numpy as np

from toqito.states import brauer


def test_brauer_2_dim_2_pval():
    """Generate Brauer states on 4 qubits."""
    expected_res = np.array(
        [
            [1, 1, 1],
            [0, 0, 0],
            [0, 0, 0],
            [1, 0, 0],
            [0, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 1],
            [0, 1, 0],
            [0, 0, 0],
            [1, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [1, 1, 1],
        ]
    )

    res = brauer(2, 2)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Is matrix a projection matrix."""
import numpy as np
from toqito.matrix_props import is_positive_semidefinite, is_square


def is_projection(mat: np.ndarray, rtol: float = 1e-05, atol: float = 1e-08) -> bool:
    r"""
    Check if matrix is a projection matrix [WikProj]_.

    A matrix is a projection matrix if it is positive semidefinite (PSD) and if

    .. math::
        \begin{equation}
            X^2 = X
        \end{equation}

    where :math:`X` is the matrix in question.

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                0 & 1 \\
                0 & 1
            \end{pmatrix}

    our function indicates that this is indeed a projection matrix.

    >>> from toqito.matrix_props import is_projection
    >>> import numpy as np
    >>> A = np.array([[0, 1], [0, 1]])
    >>> is_projection(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                -1 & -1 \\
                -1 & -1
            \end{pmatrix}

    is not positive definite.

    >>> from toqito.matrix_props import is_projection
    >>> import numpy as np
    >>> B = np.array([[-1, -1], [-1, -1]])
    >>> is_projection(B)
    False

    References
    ==========
    .. [WikProj] Wikipedia: Projection matrix.
        https://en.wikipedia.org/wiki/Projection_matrix

    :param mat: Matrix to check.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: Return :code:`True` if matrix is a projection matrix, and :code:`False` otherwise.
    """
    if not is_square(mat):
        return False

    if not is_positive_semidefinite(mat):
        return False
    return np.allclose(np.linalg.matrix_power(mat, 2), mat, rtol=rtol, atol=atol)



"""Test is_identity."""
import numpy as np

from toqito.matrix_props import is_identity


def test_is_hermitian():
    """Test if identity matrix return True."""
    mat = np.eye(4)
    np.testing.assert_equal(is_identity(mat), True)


def test_is_non_identity():
    """Test non-identity matrix."""
    mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    np.testing.assert_equal(is_identity(mat), False)


def test_is_identity_not_square():
    """Input must be a square matrix."""
    mat = np.array([[-1, 1, 1], [1, 2, 3]])
    np.testing.assert_equal(is_identity(mat), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Apply channel a subsystem of an operator."""
from typing import List, Union
import numpy as np

from toqito.channel_ops import apply_channel
from toqito.states import max_entangled
from toqito.perms import permute_systems


def partial_channel(
    rho: np.ndarray,
    phi_map: Union[np.ndarray, List[List[np.ndarray]]],
    sys: int = 2,
    dim: Union[List[int], np.ndarray] = None,
) -> np.ndarray:
    r"""Apply channel to a subsystem of an operator [WatPMap18]_.

    Applies the operator

    .. math::
        \left(\mathbb{I} \otimes \Phi \right) \left(\rho \right).

    In other words, it is the result of applying the channel :math:`\Phi` to the second subsystem
    of :math:`\rho`, which is assumed to act on two subsystems of equal dimension.

    The input :code:`phi_map` should be provided as a Choi matrix.

    This function is adapted from the QETLAB package.

    Examples
    ==========

    >>> from toqito.channel_ops import partial_channel
    >>> from toqito.channels import depolarizing
    >>> rho = np.array([[0.3101, -0.0220-0.0219*1j, -0.0671-0.0030*1j, -0.0170-0.0694*1j],
    >>>                 [-0.0220+0.0219*1j, 0.1008, -0.0775+0.0492*1j, -0.0613+0.0529*1j],
    >>>                 [-0.0671+0.0030*1j, -0.0775-0.0492*1j, 0.1361, 0.0602 + 0.0062*1j],
    >>>                 [-0.0170+0.0694*1j, -0.0613-0.0529*1j, 0.0602-0.0062*1j, 0.4530]])
    >>> phi_x = partial_channel(rho, depolarizing(2))
    [[ 0.20545+0.j       0.     +0.j      -0.0642 +0.02495j  0.     +0.j     ]
     [ 0.     +0.j       0.20545+0.j       0.     +0.j      -0.0642 +0.02495j]
     [-0.0642 -0.02495j  0.     +0.j       0.29455+0.j       0.     +0.j     ]
     [ 0.     +0.j      -0.0642 -0.02495j  0.     +0.j       0.29455+0.j     ]]

    >>> from toqito.channel_ops import partial_channel
    >>> from toqito.channels import depolarizing
    >>> rho = np.array([[0.3101, -0.0220-0.0219*1j, -0.0671-0.0030*1j, -0.0170-0.0694*1j],
    >>>                 [-0.0220+0.0219*1j, 0.1008, -0.0775+0.0492*1j, -0.0613+0.0529*1j],
    >>>                 [-0.0671+0.0030*1j, -0.0775-0.0492*1j, 0.1361, 0.0602 + 0.0062*1j],
    >>>                 [-0.0170+0.0694*1j, -0.0613-0.0529*1j, 0.0602-0.0062*1j, 0.4530]])
    >>> phi_x = partial_channel(rho, depolarizing(2), 1)
    [[0.2231+0.j      0.0191-0.00785j 0.    +0.j      0.    +0.j     ]
     [0.0191+0.00785j 0.2769+0.j      0.    +0.j      0.    +0.j     ]
     [0.    +0.j      0.    +0.j      0.2231+0.j      0.0191-0.00785j]
     [0.    +0.j      0.    +0.j      0.0191+0.00785j 0.2769+0.j     ]]

    References
    ==========
    .. [WatPMap18] Watrous, John.
        The theory of quantum information.
        Cambridge University Press, 2018.

    :param rho: A matrix.
    :param phi_map: The map to partially apply.
    :param sys: Scalar or vector specifying the size of the subsystems.
    :param dim: Dimension of the subsystems. If :code:`None`, all dimensions
                are assumed to be equal.
    :return: The partial map :code:`phi_map` applied to matrix :code:`rho`.
    """
    if dim is None:
        dim = np.round(np.sqrt(list(rho.shape))).conj().T * np.ones(2)
    if isinstance(dim, list):
        dim = np.array(dim)

    # Force dim to be a row vector.
    if dim.ndim == 1:
        dim = dim.T.flatten()
        dim = np.array([dim, dim])

    prod_dim_r1 = int(np.prod(dim[0, : sys - 1]))
    prod_dim_c1 = int(np.prod(dim[1, : sys - 1]))
    prod_dim_r2 = int(np.prod(dim[0, sys:]))
    prod_dim_c2 = int(np.prod(dim[1, sys:]))

    if isinstance(phi_map, list):
        # Compute the Kraus operators on the full system.
        s_phi_1, s_phi_2 = len(phi_map), len(phi_map[0])

        # Map is completely positive.
        if s_phi_2 == 1 or s_phi_1 == 1 and s_phi_2 > 2:
            phi = []
            for i, _ in enumerate(phi_map):
                phi.append(
                    np.kron(
                        np.kron(np.identity(prod_dim_r1), phi_map[i]),
                        np.identity(prod_dim_r2),
                    )
                )
            phi_x = apply_channel(rho, phi)
        else:
            phi_1 = []
            for i, _ in enumerate(phi_map):
                phi_1.append(
                    np.kron(
                        np.kron(np.identity(prod_dim_r1), phi_map[i][0]),
                        np.identity(prod_dim_r2),
                    )
                )
            phi_2 = []
            for i, _ in enumerate(phi_map):
                phi_2.append(
                    np.kron(
                        np.kron(np.identity(prod_dim_c1), phi_map[i][1]),
                        np.identity(prod_dim_c2),
                    )
                )

            phi_x = [list(l) for l in zip(phi_1, phi_2)]
            phi_x = apply_channel(rho, phi_x)
        return phi_x

    # The `phi_map` variable is provided as a Choi matrix.
    if isinstance(phi_map, np.ndarray):
        dim_phi = phi_map.shape

        dim = np.array(
            [
                [
                    prod_dim_r2,
                    prod_dim_r2,
                    int(dim[0, sys - 1]),
                    int(dim_phi[0] / dim[0, sys - 1]),
                    prod_dim_r1,
                    prod_dim_r1,
                ],
                [
                    prod_dim_c2,
                    prod_dim_c2,
                    int(dim[1, sys - 1]),
                    int(dim_phi[1] / dim[1, sys - 1]),
                    prod_dim_c1,
                    prod_dim_c1,
                ],
            ]
        )
        psi_r1 = max_entangled(prod_dim_r2, False, False)
        psi_c1 = max_entangled(prod_dim_c2, False, False)
        psi_r2 = max_entangled(prod_dim_r1, False, False)
        psi_c2 = max_entangled(prod_dim_c1, False, False)

        phi_map = permute_systems(
            np.kron(np.kron(psi_r1 * psi_c1.conj().T, phi_map), psi_r2 * psi_c2.conj().T),
            [1, 3, 5, 2, 4, 6],
            dim,
        )

        phi_x = apply_channel(rho, phi_map)

        return phi_x

    raise ValueError(
        "The `phi_map` variable is assumed to be provided as "
        "either a Choi matrix or a list of Kraus operators."
    )



"""Tests for NonlocalGame class."""
import unittest
import numpy as np

from toqito.nonlocal_games.nonlocal_game import NonlocalGame


class TestNonlocalGame(unittest.TestCase):
    """Unit test for NonlocalGame."""

    @staticmethod
    def ffl_nonlocal_game():
        """Define the FFL nonlocal game."""
        num_alice_in, num_alice_out = 2, 2
        num_bob_in, num_bob_out = 2, 2
        prob_mat = np.array([[1 / 3, 1 / 3], [1 / 3, 0]])

        pred_mat = np.zeros((num_alice_out, num_bob_out, num_alice_in, num_bob_in))
        for a_alice in range(num_alice_out):
            for b_bob in range(num_bob_out):
                for x_alice in range(num_alice_in):
                    for y_bob in range(num_bob_in):
                        if (a_alice or x_alice) != (b_bob or y_bob):
                            pred_mat[a_alice, b_bob, x_alice, y_bob] = 1
        return prob_mat, pred_mat

    @staticmethod
    def chsh_nonlocal_game():
        """Define the CHSH nonlocal game."""
        num_alice_inputs, num_alice_outputs = 2, 2
        num_bob_inputs, num_bob_outputs = 2, 2
        prob_mat = np.array([[1 / 4, 1 / 4], [1 / 4, 1 / 4]])
        pred_mat = np.zeros((num_alice_outputs, num_bob_outputs, num_alice_inputs, num_bob_inputs))

        for a_alice in range(num_alice_outputs):
            for b_bob in range(num_bob_outputs):
                for x_alice in range(num_alice_inputs):
                    for y_bob in range(num_bob_inputs):
                        if a_alice ^ b_bob == x_alice * y_bob:
                            pred_mat[a_alice, b_bob, x_alice, y_bob] = 1
        return prob_mat, pred_mat

    def test_chsh_lower_bound(self):
        """Calculate the lower bound on the quantum value for the CHSH game."""
        prob_mat, pred_mat = self.chsh_nonlocal_game()
        chsh = NonlocalGame(prob_mat, pred_mat)
        res = chsh.quantum_value_lower_bound()

        # It may be possible for the lower bound to not be equal to the correct
        # quantum value (as it may get stuck in a local minimum), but it should
        # never be possible for the lower bound to attain a value higher than
        # the true quantum value.
        self.assertLessEqual(np.isclose(res, np.cos(np.pi / 8) ** 2, rtol=1e-02), True)

    def test_chsh_game_classical_value(self):
        """Classical value for the CHSH game."""
        prob_mat, pred_mat = self.chsh_nonlocal_game()

        chsh = NonlocalGame(prob_mat, pred_mat)
        res = chsh.classical_value()
        expected_res = 3 / 4
        self.assertEqual(np.isclose(res, expected_res), True)

    def test_chsh_game_classical_value_rep_2(self):
        r"""
        Classical value for the CHSH game for 2 reps.

        Note that for classical strategies, it is known that parallel repetition
        does *not* hold for the CHSH game, that is:

        w_c(CHSH \land CHSH) = 10/16 > 9/16 = w_c(CHSH) w_c(CHSH).
        """
        prob_mat, pred_mat = self.chsh_nonlocal_game()

        chsh = NonlocalGame(prob_mat, pred_mat, 2)
        res = chsh.classical_value()
        expected_res = 10 / 16
        self.assertEqual(np.isclose(res, expected_res), True)

    def test_ffl_game_classical_value(self):
        """Classical value for the FFL game."""
        prob_mat, pred_mat = self.ffl_nonlocal_game()

        ffl = NonlocalGame(prob_mat, pred_mat)
        res = ffl.classical_value()
        expected_res = 2 / 3
        self.assertEqual(np.isclose(res, expected_res), True)

    def test_ffl_game_classical_value_rep_2(self):
        """Classical value for the FFL game for 2 reps."""
        prob_mat, pred_mat = self.ffl_nonlocal_game()

        ffl = NonlocalGame(prob_mat, pred_mat, 2)
        res = ffl.classical_value()
        expected_res = 2 / 3
        self.assertEqual(np.isclose(res, expected_res), True)

    def test_ffl_game_quantum_lower_bound_value_rep_2(self):
        """Lower bound on quantum value for the FFL game for 2 reps."""
        prob_mat, pred_mat = self.ffl_nonlocal_game()

        ffl = NonlocalGame(prob_mat, pred_mat, 2)
        res = ffl.quantum_value_lower_bound()
        expected_res = 2 / 3
        self.assertLessEqual(np.isclose(res, expected_res), True)

    def test_ffl_game_nonsignaling_value(self):
        """Non-signaling value for the FFL game."""
        prob_mat, pred_mat = self.ffl_nonlocal_game()

        ffl = NonlocalGame(prob_mat, pred_mat)
        res = ffl.nonsignaling_value()
        expected_res = 2 / 3
        self.assertEqual(np.isclose(res, expected_res, atol=0.5), True)

    def test_ffl_game_nonsignaling_value_rep_2(self):
        """Non-signaling value for the FFL game for 2 reps."""
        prob_mat, pred_mat = self.ffl_nonlocal_game()

        ffl = NonlocalGame(prob_mat, pred_mat, 2)
        res = ffl.nonsignaling_value()
        expected_res = 2 / 3
        self.assertEqual(np.isclose(res, expected_res, atol=0.5), True)

    def test_chsh_game_nonsignaling_value(self):
        """Non-signaling value for the CHSH game."""
        prob_mat, pred_mat = self.chsh_nonlocal_game()

        chsh = NonlocalGame(prob_mat, pred_mat)
        res = chsh.nonsignaling_value()
        expected_res = 1
        self.assertEqual(np.isclose(res, expected_res, atol=0.5), True)

    def test_chsh_game_nonsignaling_value_rep_2(self):
        """Non-signaling value for the CHSH game for 2 reps."""
        prob_mat, pred_mat = self.chsh_nonlocal_game()

        chsh = NonlocalGame(prob_mat, pred_mat, 2)
        res = chsh.nonsignaling_value()
        expected_res = 1
        self.assertEqual(np.isclose(res, expected_res, atol=0.5), True)


if __name__ == "__main__":
    unittest.main()



"""Test is_product_vector."""
import numpy as np

from toqito.state_props import is_product_vector
from toqito.states import basis
from toqito.states import max_entangled


def test_is_product_entangled_state():
    """Check that is_product_vector returns False for an entangled state."""
    ent_vec = max_entangled(3)
    np.testing.assert_equal(is_product_vector(ent_vec), False)


def test_is_product_entangled_state_2_sys():
    """Check that dimension argument as list is supported."""
    ent_vec = max_entangled(4)
    np.testing.assert_equal(is_product_vector(ent_vec, dim=[4, 4]), False)


def test_is_product_entangled_state_3_sys():
    """Check that dimension argument as list is supported."""
    ent_vec = max_entangled(4)
    np.testing.assert_equal(is_product_vector(ent_vec, dim=[2, 2, 2, 2]), False)


def test_is_product_separable_state():
    """Check that is_product_vector returns True for a separable state."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    sep_vec = (
        1 / 2 * (np.kron(e_0, e_0) - np.kron(e_0, e_1) - np.kron(e_1, e_0) + np.kron(e_1, e_1))
    )
    np.testing.assert_equal(is_product_vector(sep_vec), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Concurrence property."""
import numpy as np

from toqito.matrices import pauli


def concurrence(rho: np.ndarray) -> float:
    r"""
    Calculate the concurrence of a bipartite state [WikCon]_.

    The concurrence of a bipartite state :math:`\rho` is defined as

    .. math::
        \max(0, \lambda_1 - \lambda_2 - \lambda_3 - \lambda_4),

    where :math:`\lambda_1, \ldots, \lambda_4` are the eigenvalues in decreasing order of the
    matrix.

    Concurrence can serve as a measure of entanglement.

    Examples
    ==========

    Consider the following Bell state:

    .. math::
        u = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right).

    The concurrence of the density matrix :math:`\rho = u u^*` defined by the vector :math:`u` is
    given as

    .. math::
        \mathcal{C}(\rho) \approx 1.

    The following example calculates this quantity using the :code:`toqito` package.

    >>> import numpy as np
    >>> from toqito.states import basis
    >>> from toqito.state_props import concurrence
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> e_00, e_11 = np.kron(e_0, e_0), np.kron(e_1, e_1)
    >>> u_vec = 1 / np.sqrt(2) * (e_00 + e_11)
    >>> rho = u_vec * u_vec.conj().T
    >>> concurrence(rho)
    0.9999999999999998

    Consider the concurrence of the following product state

    .. math::
        v = |0\rangle \otimes |1 \rangle.

    As this state has no entanglement, the concurrence is zero.

    >>> import numpy as np
    >>> from toqito.states import basis
    >>> from toqito.state_props import concurrence
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> v_vec = np.kron(e_0, e_1)
    >>> sigma = v_vec * v_vec.conj().T
    >>> concurrence(sigma)
    0

    References
    ==========
    .. [WikCon] Wikipedia page for concurrence (quantum computing)
       https://en.wikipedia.org/wiki/Concurrence_(quantum_computing)

    :param rho: The bipartite system specified as a matrix.
    :return: The concurrence of the bipartite state :math:`\rho`.
    """
    if rho.shape != (4, 4):
        raise ValueError("InvalidDim: Concurrence is only defined for bipartite systems.")

    sigma_y = pauli("Y", False)
    sigma_y_y = np.kron(sigma_y, sigma_y)

    rho_hat = sigma_y_y @ rho.conj().T @ sigma_y_y

    eig_vals = np.linalg.eigvals(rho @ rho_hat)
    eig_vals = np.sort(np.abs(np.sqrt(eig_vals)))[::-1]
    return max(0, eig_vals[0] - eig_vals[1] - eig_vals[2] - eig_vals[3])



"""Domino state."""
import numpy as np

from toqito.states import basis


def domino(idx: int) -> np.ndarray:
    r"""
    Produce a domino state [CBDOM99]_, [UPB99]_.

    The orthonormal product basis of domino states is given as

    .. math::
        \begin{equation}
            \begin{aligned}
            |\phi_0\rangle = |1\rangle |1 \rangle,
            \qquad
            |\phi_1\rangle = |0 \rangle \left(\frac{|0 \rangle + |1 \rangle}{\sqrt{2}} \right),
            & \qquad
            |\phi_2\rangle = |0\rangle \left(\frac{|0\rangle - |1\rangle}{\sqrt{2}}\right), \\
            |\phi_3\rangle = |2\rangle \left(\frac{|0\rangle + |1\rangle}{\sqrt{2}}\right), \qquad
            |\phi_4\rangle = |2\rangle \left(\frac{|0\rangle - |1\rangle}{\sqrt{2}}\right), & \qquad
            |\phi_5\rangle = \left(\frac{|0\rangle + |1\rangle}{\sqrt{2}}\right) |0\rangle, \\
            |\phi_6\rangle = \left(\frac{|0\rangle - |1\rangle}{\sqrt{2}}\right) |0\rangle, \qquad
            |\phi_7\rangle = \left(\frac{|0\rangle + |1\rangle}{\sqrt{2}}\right) |2\rangle, & \qquad
            |\phi_8\rangle = \left(\frac{|0\rangle - |1\rangle}{\sqrt{2}}\right) |2\rangle.
            \end{aligned}
        \end{equation}

    Returns one of the following nine domino states depending on the value of :code:`idx`.

    Examples
    ==========

    When :code:`idx = 0`, this produces the following Domino state

    .. math::
        |\phi_0 \rangle = |11 \rangle |11 \rangle.

    Using :code:`toqito`, we can see that this yields the proper state.

    >>> from toqito.states import domino
    >>> domino(0)
    [[0],
     [0],
     [0],
     [0],
     [1],
     [0],
     [0],
     [0],
     [0]]

    When :code:`idx = 3`, this produces the following Domino state

    .. math::
        |\phi_3\rangle = |2\rangle \left(\frac{|0\rangle + |1\rangle}
        {\sqrt{2}}\right)

    Using :code:`toqito`, we can see that this yields the proper state.

    >>> from toqito.states import domino
    >>> domino(3)
    [[0.        ],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.70710678],
     [0.70710678]]

    References
    ==========
    .. [CBDOM99] Bennett, Charles H., et al.
        Quantum nonlocality without entanglement.
        Phys. Rev. A, 59:1070–1091, Feb 1999.
        https://arxiv.org/abs/quant-ph/9804053

    .. [UPB99] Bennett, Charles H., et al.
        "Unextendible product bases and bound entanglement."
        Physical Review Letters 82.26 (1999): 5385.
        https://arxiv.org/abs/quant-ph/9808030

    :param idx: A parameter in [0, 1, 2, 3, 4, 5, 6, 7, 8]
    :return: Domino state of index :code:`idx`.
    """
    e_0, e_1, e_2 = basis(3, 0), basis(3, 1), basis(3, 2)
    if idx == 0:
        return np.kron(e_1, e_1)
    if idx == 1:
        return np.kron(e_0, 1 / np.sqrt(2) * (e_0 + e_1))
    if idx == 2:
        return np.kron(e_0, 1 / np.sqrt(2) * (e_0 - e_1))
    if idx == 3:
        return np.kron(e_2, 1 / np.sqrt(2) * (e_1 + e_2))
    if idx == 4:
        return np.kron(e_2, 1 / np.sqrt(2) * (e_1 - e_2))
    if idx == 5:
        return np.kron(1 / np.sqrt(2) * (e_1 + e_2), e_0)
    if idx == 6:
        return np.kron(1 / np.sqrt(2) * (e_1 - e_2), e_0)
    if idx == 7:
        return np.kron(1 / np.sqrt(2) * (e_0 + e_1), e_2)
    if idx == 8:
        return np.kron(1 / np.sqrt(2) * (e_0 - e_1), e_2)
    raise ValueError("Invalid integer value for Domino state.")



"""Is matrix a density matrix."""
import numpy as np
from toqito.matrix_props import is_positive_semidefinite


def is_density(mat: np.ndarray) -> bool:
    r"""
    Check if matrix is a density matrix [WikDensity]_.

    A matrix is a density matrix if its trace is equal to one and it has the
    property of being positive semidefinite (PSD).

    Examples
    ==========

    Consider the Bell state:

    .. math::
       u = \frac{1}{\sqrt{2}} |00 \rangle + \frac{1}{\sqrt{2}} |11 \rangle.

    Constructing the matrix :math:`\rho = u u^*` defined as

    .. math::
        \rho = \frac{1}{2} \begin{pmatrix}
                                1 & 0 & 0 & 1 \\
                                0 & 0 & 0 & 0 \\
                                0 & 0 & 0 & 0 \\
                                1 & 0 & 0 & 1
                           \end{pmatrix}

    our function indicates that this is indeed a density operator as the trace
    of :math:`\rho` is equal to :math:`1` and the matrix is positive
    semidefinite:

    >>> from toqito.matrix_props import is_density
    >>> from toqito.states import bell
    >>> import numpy as np
    >>> rho = bell(0) * bell(0).conj().T
    >>> is_density(rho)
    True

    Alternatively, the following example matrix :math:`\sigma` defined as

    .. math::
        \sigma = \frac{1}{2} \begin{pmatrix}
                                1 & 2 \\
                                3 & 1
                             \end{pmatrix}

    does satisfy :math:`\text{Tr}(\sigma) = 1`, however fails to be positive
    semidefinite, and is therefore not a density operator. This can be
    illustrated using :code:`toqito` as follows.

    >>> from toqito.matrix_props import is_density
    >>> from toqito.states import bell
    >>> import numpy as np
    >>> sigma = 1/2 * np.array([[1, 2], [3, 1]])
    >>> is_density(sigma)
    False

    References
    ==========
    .. [WikDensity] Wikipedia: Density matrix
        https://en.wikipedia.org/wiki/Density_matrix

    :param mat: Matrix to check.
    :return: Return `True` if matrix is a density matrix, and `False`
             otherwise.
    """
    return is_positive_semidefinite(mat) and np.isclose(np.trace(mat), 1)



"""Test majorizes."""
import numpy as np

from toqito.channels import partial_trace
from toqito.matrix_props import majorizes
from toqito.states import max_entangled


def test_majorizes_simple_example():
    """Test that simple example of vectors returns True."""
    np.testing.assert_equal(majorizes([3, 0, 0], [1, 1, 1]), True)


def test_majorizes_max_entangled():
    """Test that max entangled partial trace returns False."""
    v_vec = max_entangled(3)
    rho = v_vec * v_vec.conj().T
    np.testing.assert_equal(majorizes(partial_trace(rho), rho), False)


def test_majorizes_max_entangled_flip():
    """Test that max entangled partial trace returns True (flipped args)."""
    v_vec = max_entangled(3)
    rho = v_vec * v_vec.conj().T
    np.testing.assert_equal(majorizes(rho, partial_trace(rho)), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test is_normal."""
import numpy as np

from toqito.matrix_props import is_normal


def test_is_normal():
    """Check that normal and non-unitary and non-Hermitian matrix yields True.

    Normal matrix obtained from:
    https://en.wikipedia.org/wiki/Normal_matrix
    """
    mat = np.array([[1, 1, 0], [0, 1, 1], [1, 0, 1]])
    np.testing.assert_equal(is_normal(mat), True)


def test_is_normal_identity():
    """Test that the identity matrix returns True."""
    mat = np.identity(4)
    np.testing.assert_equal(is_normal(mat), True)


def test_is_not_normal():
    """Test that non-normal matrix returns False."""
    mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    np.testing.assert_equal(is_normal(mat), False)


def test_is_normal_not_square():
    """Input must be a square matrix."""
    mat = np.array([[-1, 1, 1], [1, 2, 3]])
    np.testing.assert_equal(is_normal(mat), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test dephasing."""
import numpy as np

from toqito.channels import dephasing
from toqito.channel_ops import apply_channel


def test_dephasing_completely_dephasing():
    """The completely dephasing channel kills everything off diagonal."""
    test_input_mat = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])

    expected_res = np.array([[1, 0, 0, 0], [0, 6, 0, 0], [0, 0, 11, 0], [0, 0, 0, 16]])

    res = apply_channel(test_input_mat, dephasing(4))

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_dephasing_partially_dephasing():
    """The partially dephasing channel for `p = 0.5`."""
    test_input_mat = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]])

    expected_res = np.array([[17.5, 0, 0, 0], [0, 20, 0, 0], [0, 0, 22.5, 0], [0, 0, 0, 25]])

    res = apply_channel(test_input_mat, dephasing(4, 0.5))

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Helstrom-Holevo metric."""
import numpy as np

from toqito.matrix_props import is_density
from toqito.state_metrics import trace_norm


def helstrom_holevo(rho: np.ndarray, sigma: np.ndarray) -> float:
    r"""
    Compute the Helstrom-Holevo distance between density matrices [WikHeHo]_.

    In general, the best success probability to discriminate two mixed states represented by
    :math:`\rho` and :math:`\sigma` is given by [WikHeHo]_.

    .. math::
         \frac{1}{2}+\frac{1}{2} \left(\frac{1}{2} \left|\rho - \sigma \right|_1\right).

    Examples
    ==========
    Consider the following Bell state

    .. math::
        u = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right) \in \mathcal{X}.

    The corresponding density matrix of :math:`u` may be calculated by:

    .. math::
        \rho = u u^* = \begin{pmatrix}
                         1 & 0 & 0 & 1 \\
                         0 & 0 & 0 & 0 \\
                         0 & 0 & 0 & 0 \\
                         1 & 0 & 0 & 1
                       \end{pmatrix} \in \text{D}(\mathcal{X}).

    Calculating the Helstrom-Holevo distance of states that are identical yield a value of
    :math:`1/2`. This can be verified in :code:`toqito` as follows.

    >>> from toqito.states import basis
    >>> from toqito.state_metrics import helstrom_holevo
    >>> import numpy as np
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> e_00 = np.kron(e_0, e_0)
    >>> e_11 = np.kron(e_1, e_1)
    >>>
    >>> u_vec = 1 / np.sqrt(2) * (e_00 + e_11)
    >>> rho = u_vec * u_vec.conj().T
    >>> sigma = rho
    >>>
    >>> helstrom_holevo(rho, sigma)
    0.5

    References
    ==========
    .. [WikHeHo] Wikipedia: Holevo's theorem.
        https://en.wikipedia.org/wiki/Holevo%27s_theorem

    :param rho: Density operator.
    :param sigma: Density operator.
    :return: The Helstrom-Holevo distance between :code:`rho` and :code:`sigma`.
    """
    if not is_density(rho) or not is_density(sigma):
        raise ValueError("Helstrom-Holevo is only defined for density operators.")
    return 1 / 2 + 1 / 2 * (trace_norm(rho - sigma)) / 2



"""Test measure."""
import numpy as np

from toqito.measurement_ops import measure
from toqito.states import basis


def test_measure_state():
    """Test measure on quantum state."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    psi = 1 / np.sqrt(3) * e_0 + np.sqrt(2 / 3) * e_1
    rho = psi * psi.conj().T

    proj_0 = e_0 * e_0.conj().T
    proj_1 = e_1 * e_1.conj().T
    np.testing.assert_equal(np.isclose(measure(proj_0, rho), 1 / 3), True)
    np.testing.assert_equal(np.isclose(measure(proj_1, rho), 2 / 3), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test is_ppt."""
import numpy as np

from toqito.state_props import is_ppt
from toqito.states import bell, horodecki


def test_is_ppt():
    """Check that PPT matrix returns True."""
    mat = np.identity(9)
    np.testing.assert_equal(is_ppt(mat), True)


def test_is_ppt_sys():
    """Check that PPT matrix returns True with sys specified."""
    mat = np.identity(9)
    np.testing.assert_equal(is_ppt(mat, 2), True)


def test_is_ppt_dim_sys():
    """Check that PPT matrix returns True with dim and sys specified."""
    mat = np.identity(9)
    np.testing.assert_equal(is_ppt(mat, 2, np.round(np.sqrt(mat.size))), True)


def test_is_ppt_tol():
    """Check that PPT matrix returns True."""
    mat = np.identity(9)
    np.testing.assert_equal(is_ppt(mat, 2, np.round(np.sqrt(mat.size)), 1e-10), True)


def test_entangled_state():
    """Entangled state of dimension 2 will violate PPT criterion."""
    rho = bell(2) * bell(2).conj().T
    np.testing.assert_equal(is_ppt(rho), False)


def test_is_horodecki_ppt():
    """Horodecki state is an example of an entangled PPT state."""
    np.testing.assert_equal(is_ppt(horodecki(0.5, [3, 3])), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Tests for channel_fidelity."""
import numpy as np

from toqito.channels import dephasing, depolarizing
from toqito.channel_metrics import channel_fidelity


def test_channel_fidelity_same_channel():
    """The fidelity of identical channels should yield 1."""
    choi_1 = dephasing(4)
    choi_2 = dephasing(4)
    np.testing.assert_equal(np.isclose(channel_fidelity(choi_1, choi_2), 1, atol=1e-3), True)


def test_channel_fidelity_different_channel():
    """Calculate the channel fidelity of different channels."""
    choi_1 = dephasing(4)
    choi_2 = depolarizing(4)
    np.testing.assert_equal(np.isclose(channel_fidelity(choi_1, choi_2), 1 / 2, atol=1e-3), True)


def test_channel_fidelity_inconsistent_dims():
    """Inconsistent dimensions between Choi matrices."""
    with np.testing.assert_raises(ValueError):
        choi_1 = depolarizing(4)
        choi_2 = dephasing(2)
        channel_fidelity(choi_1, choi_2)


def test_channel_fidelity_non_square():
    """Non-square inputs for channel fidelity."""
    with np.testing.assert_raises(ValueError):
        choi_1 = np.array([[1, 2, 3], [4, 5, 6]])
        choi_2 = np.array([[1, 2, 3], [4, 5, 6]])
        channel_fidelity(choi_1, choi_2)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Tests for kraus_to_choi."""
import numpy as np

from toqito.channels import dephasing, depolarizing
from toqito.channel_ops import kraus_to_choi


def test_kraus_to_choi_swap_operator():
    """Choi matrix of the transpose map is the swap operator."""
    kraus_1 = np.array([[1, 0], [0, 0]])
    kraus_2 = np.array([[1, 0], [0, 0]]).conj().T
    kraus_3 = np.array([[0, 1], [0, 0]])
    kraus_4 = np.array([[0, 1], [0, 0]]).conj().T
    kraus_5 = np.array([[0, 0], [1, 0]])
    kraus_6 = np.array([[0, 0], [1, 0]]).conj().T
    kraus_7 = np.array([[0, 0], [0, 1]])
    kraus_8 = np.array([[0, 0], [0, 1]]).conj().T

    kraus_ops = [
        [kraus_1, kraus_2],
        [kraus_3, kraus_4],
        [kraus_5, kraus_6],
        [kraus_7, kraus_8],
    ]

    choi_res = kraus_to_choi(kraus_ops)
    expected_choi_res = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    bool_mat = np.isclose(choi_res, expected_choi_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_kraus_to_choi_swap_operator_non_unique():
    """As Kraus operators are non-unique, these also should yield the swap operator."""
    kraus_1 = np.array([[0, 1 / np.sqrt(2)], [1 / np.sqrt(2), 0]])
    kraus_2 = np.array([[0, 1 / np.sqrt(2)], [1 / np.sqrt(2), 0]]).conj().T
    kraus_3 = np.array([[1, 0], [0, 0]])
    kraus_4 = np.array([[1, 0], [0, 0]]).conj().T
    kraus_5 = np.array([[0, 0], [0, 1]])
    kraus_6 = np.array([[0, 0], [0, 1]]).conj().T
    kraus_7 = np.array([[0, 1 / np.sqrt(2)], [-1 / np.sqrt(2), 0]])
    kraus_8 = np.array([[0, 1 / np.sqrt(2)], [-1 / np.sqrt(2), 0]]).conj().T

    kraus_ops = [
        [kraus_1, kraus_2],
        [kraus_3, kraus_4],
        [kraus_5, kraus_6],
        [kraus_7, kraus_8],
    ]

    choi_res = kraus_to_choi(kraus_ops)
    expected_choi_res = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    bool_mat = np.isclose(choi_res, expected_choi_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_kraus_to_choi_dephasing_channel():
    """Kraus operators for dephasing channel should yield the proper Choi matrix."""
    kraus_1 = np.array([[1, 0], [0, 0]])
    kraus_2 = np.array([[1, 0], [0, 0]])
    kraus_3 = np.array([[0, 0], [0, 1]])
    kraus_4 = np.array([[0, 0], [0, 1]])

    kraus_ops = [[kraus_1, kraus_2], [kraus_3, kraus_4]]

    choi_res = kraus_to_choi(kraus_ops)
    expected_choi_res = dephasing(2)

    bool_mat = np.isclose(choi_res, expected_choi_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_kraus_to_choi_depolarizing_channel():
    """Kraus operators for depolarizing channel should yield the proper Choi matrix."""
    kraus_1 = np.array([[1 / np.sqrt(2), 0], [0, 0]])
    kraus_2 = np.array([[1 / np.sqrt(2), 0], [0, 0]])
    kraus_3 = np.array([[0, 0], [1 / np.sqrt(2), 0]])
    kraus_4 = np.array([[0, 0], [1 / np.sqrt(2), 0]])
    kraus_5 = np.array([[0, 1 / np.sqrt(2)], [0, 0]])
    kraus_6 = np.array([[0, 1 / np.sqrt(2)], [0, 0]])
    kraus_7 = np.array([[0, 0], [0, 1 / np.sqrt(2)]])
    kraus_8 = np.array([[0, 0], [0, 1 / np.sqrt(2)]])

    kraus_ops = [
        [kraus_1, kraus_2],
        [kraus_3, kraus_4],
        [kraus_5, kraus_6],
        [kraus_7, kraus_8],
    ]

    choi_res = kraus_to_choi(kraus_ops)
    expected_choi_res = depolarizing(2)

    bool_mat = np.isclose(choi_res, expected_choi_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test schmidt_rank."""
import numpy as np

from toqito.state_props import schmidt_rank
from toqito.states import basis
from toqito.states import bell


def test_schmidt_rank_bell_state():
    """
    Computing the Schmidt rank of the entangled Bell state should yield a
    value greater than 1.
    """
    np.testing.assert_equal(np.isclose(schmidt_rank(bell(0)), 2), True)


def test_schmidt_rank_bell_state_dim_1():
    """
    Computing the Schmidt rank of Bell state with dim 1.
    """
    np.testing.assert_equal(np.isclose(schmidt_rank(bell(0), 1), 1), True)


def test_schmidt_rank_bell_state_list():
    """
    Computing the Schmidt rank of Bell state with list as argument for dims.
    """
    np.testing.assert_equal(np.isclose(schmidt_rank(bell(0), [2, 2]), 2), True)


def test_schmidt_rank_singlet_state():
    """
    Computing the Schmidt rank of the entangled singlet state should yield
    a value greater than 1.
    """
    e_0, e_1 = basis(2, 0), basis(2, 1)
    rho = 1 / np.sqrt(2) * (np.kron(e_0, e_1) - np.kron(e_1, e_0))
    np.testing.assert_equal(schmidt_rank(rho) > 1, True)


def test_schmidt_rank_separable_state():
    """
    Computing the Schmidt rank of a separable state should yield a value
    equal to 1.
    """
    e_0, e_1 = basis(2, 0), basis(2, 1)
    e_00 = np.kron(e_0, e_0)
    e_01 = np.kron(e_0, e_1)
    e_10 = np.kron(e_1, e_0)
    e_11 = np.kron(e_1, e_1)
    rho = 1 / 2 * (e_00 - e_01 - e_10 + e_11)
    np.testing.assert_equal(schmidt_rank(rho) == 1, True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Two-player extended nonlocal game."""
from collections import defaultdict
from typing import Dict, Tuple

import cvxpy
import numpy as np

from toqito.matrix_ops import tensor
from toqito.random import random_unitary
from toqito.helper import update_odometer


class ExtendedNonlocalGame:
    r"""
    Create two-player extended nonlocal game object.

    *Extended nonlocal games* are a superset of nonlocal games in which the
    players share a tripartite state with the referee. In such games, the
    winning conditions for Alice and Bob may depend on outcomes of measurements
    made by the referee, on its part of the shared quantum state, in addition
    to Alice and Bob's answers to the questions sent by the referee.

    Extended nonlocal games were initially defined in [JMRW16]_ and more
    information on these games can be found in [Russo17]_.

    References
    ==========
    .. [JMRW16] Nathaniel Johnston, Rajat Mittal, Vincent Russo, John Watrous
        "Extended non-local games and monogamy-of-entanglement games",
        Proceedings of the Royal Society A: Mathematical, Physical and
        Engineering Sciences 472.2189 (2016),
        https://arxiv.org/abs/1510.02083

    .. [Russo17] Vincent Russo
        "Extended nonlocal games"
        https://arxiv.org/abs/1704.07375
    """

    def __init__(self, prob_mat: np.ndarray, pred_mat: np.ndarray, reps: int = 1) -> None:
        """
        Construct extended nonlocal game object.

        :param prob_mat: A matrix whose (x, y)-entry gives the probability
                        that the referee will give Alice the value `x` and Bob
                        the value `y`.
        :param pred_mat:
        :param reps: Number of parallel repetitions to perform.
        """
        if reps == 1:
            self.prob_mat = prob_mat
            self.pred_mat = pred_mat
            self.reps = reps

        else:
            (
                dim_x,
                dim_y,
                num_alice_out,
                num_bob_out,
                num_alice_in,
                num_bob_in,
            ) = pred_mat.shape
            self.prob_mat = tensor(prob_mat, reps)

            pred_mat2 = np.zeros(
                (
                    dim_x ** reps,
                    dim_y ** reps,
                    num_alice_out ** reps,
                    num_bob_out ** reps,
                    num_alice_in ** reps,
                    num_bob_in ** reps,
                )
            )
            i_ind = np.zeros(reps, dtype=int)
            j_ind = np.zeros(reps, dtype=int)
            for i in range(num_alice_in ** reps):
                for j in range(num_bob_in ** reps):
                    to_tensor = np.empty([reps, dim_x, dim_y, num_alice_out, num_bob_out])
                    for k in range(reps - 1, -1, -1):
                        to_tensor[k] = pred_mat[:, :, :, :, i_ind[k], j_ind[k]]
                    pred_mat2[:, :, :, :, i, j] = tensor(to_tensor)
                    j_ind = update_odometer(j_ind, num_bob_in * np.ones(reps))
                i_ind = update_odometer(i_ind, num_alice_in * np.ones(reps))
            self.pred_mat = pred_mat2
            self.reps = reps

    def unentangled_value(self) -> float:
        r"""
        Calculate the unentangled value of an extended nonlocal game.

        The *unentangled value* of an extended nonlocal game is the supremum
        value for Alice and Bob's winning probability in the game over all
        unentangled strategies. Due to convexity and compactness, it is possible
        to calculate the unentangled extended nonlocal game by:

        .. math::
            \omega(G) = \max_{f, g}
            \lVert
            \sum_{(x,y) \in \Sigma_A \times \Sigma_B} \pi(x,y)
            V(f(x), g(y)|x, y)
            \rVert

        where the maximum is over all functions :math:`f : \Sigma_A \rightarrow
        \Gamma_A` and :math:`g : \Sigma_B \rightarrow \Gamma_B`.

        :return: The unentangled value of the extended nonlocal game.
        """
        dim_x, dim_y, alice_out, bob_out, alice_in, bob_in = self.pred_mat.shape

        max_unent_val = float("-inf")
        for a_out in range(alice_out):
            for b_out in range(bob_out):
                p_win = np.zeros([dim_x, dim_y], dtype=complex)
                for x_in in range(alice_in):
                    for y_in in range(bob_in):
                        p_win += (
                            self.prob_mat[x_in, y_in]
                            * self.pred_mat[:, :, a_out, b_out, x_in, y_in]
                        )

                rho = cvxpy.Variable((dim_x, dim_y), hermitian=True)

                objective = cvxpy.Maximize(cvxpy.real(cvxpy.trace(p_win.conj().T @ rho)))

                constraints = [cvxpy.trace(rho) == 1, rho >> 0]
                problem = cvxpy.Problem(objective, constraints)
                unent_val = problem.solve()
                max_unent_val = max(max_unent_val, unent_val)
        return max_unent_val

    def nonsignaling_value(self) -> float:
        r"""
        Calculate the non-signaling value of an extended nonlocal game.

        The *non-signaling value* of an extended nonlocal game is the supremum
        value of the winning probability of the game taken over all
        non-signaling strategies for Alice and Bob.

        A *non-signaling strategy* for an extended nonlocal game consists of a
        function

        .. math::
            K : \Gamma_A \times \Gamma_B \times \Sigma_A \times \Sigma_B
            \rightarrow \text{Pos}(\mathcal{R})

        such that

        .. math::
            \sum_{a \in \Gamma_A} K(a,b|x,y) = \rho_b^y
            \quad \text{and} \quad
            \sum_{b \in \Gamma_B} K(a,b|x,y) = \sigma_a^x,

        for all :math:`x \in \Sigma_A` and :math:`y \in \Sigma_B` where
        :math:`\{\rho_b^y : y \in \Sigma_A, \ b \in \Gamma_B\}` and
        :math:`\{\sigma_a^x : x \in \Sigma_A, \ a \in \Gamma_B\}` are
        collections of operators satisfying

        .. math::
            \sum_{a \in \Gamma_A} \rho_b^y =
            \tau =
            \sum_{b \in \Gamma_B} \sigma_a^x,

        for every choice of :math:`x \in \Sigma_A` and :math:`y \in \Sigma_B`
        where :math:`\tau \in \text{D}(\mathcal{R})` is a density operator.

        :return: The non-signaling value of the extended nonlocal game.
        """
        dim_x, dim_y, alice_out, bob_out, alice_in, bob_in = self.pred_mat.shape
        constraints = list()

        # The cvxpy package does not support optimizing over more than
        # 2-dimensional objects. To overcome this, we use a dictionary to index
        # between the questions and answers, while the cvxpy variables held at
        # this positions are `dim_x`-by-`dim_y` cvxpy Variable objects.

        # Define K(a,b|x,y) variable.
        k_var = defaultdict(cvxpy.Variable)
        for a_out in range(alice_out):
            for b_out in range(bob_out):
                for x_in in range(alice_in):
                    for y_in in range(bob_in):
                        k_var[a_out, b_out, x_in, y_in] = cvxpy.Variable(
                            (dim_x, dim_y), hermitian=True
                        )
                        constraints.append(k_var[a_out, b_out, x_in, y_in] >> 0)

        # Define \sigma_a^x variable.
        sigma = defaultdict(cvxpy.Variable)
        for a_out in range(alice_out):
            for x_in in range(alice_in):
                sigma[a_out, x_in] = cvxpy.Variable((dim_x, dim_y), hermitian=True)

        # Define \rho_b^y variable.
        rho = defaultdict(cvxpy.Variable)
        for b_out in range(bob_out):
            for y_in in range(bob_in):
                rho[b_out, y_in] = cvxpy.Variable((dim_x, dim_y), hermitian=True)

        # Define \tau density operator variable.
        tau = cvxpy.Variable((dim_x, dim_y), hermitian=True)

        p_win = 0
        for a_out in range(alice_out):
            for b_out in range(bob_out):
                for x_in in range(alice_in):
                    for y_in in range(bob_in):
                        p_win += self.prob_mat[x_in, y_in] * cvxpy.trace(
                            self.pred_mat[:, :, a_out, b_out, x_in, y_in].conj().T
                            @ k_var[a_out, b_out, x_in, y_in]
                        )

        objective = cvxpy.Maximize(cvxpy.real(p_win))

        # The following constraints enforce the so-called non-signaling
        # constraints.

        # Enforce that:
        # \sum_{b \in \Gamma_B} K(a,b|x,y) = \sigma_a^x
        for x_in in range(alice_in):
            for y_in in range(bob_in):
                for a_out in range(alice_out):
                    b_sum = 0
                    for b_out in range(bob_out):
                        b_sum += k_var[a_out, b_out, x_in, y_in]
                    constraints.append(b_sum == sigma[a_out, x_in])

        # Enforce non-signaling constraints on Alice marginal:
        # \sum_{a \in \Gamma_A} K(a,b|x,y) = \rho_b^y
        for x_in in range(alice_in):
            for y_in in range(bob_in):
                for b_out in range(bob_out):
                    a_sum = 0
                    for a_out in range(alice_out):
                        a_sum += k_var[a_out, b_out, x_in, y_in]
                    constraints.append(a_sum == rho[b_out, y_in])

        # Enforce non-signaling constraints on Bob marginal:
        # \sum_{a \in \Gamma_A} \sigma_a^x = \tau
        for x_in in range(alice_in):
            sig_a_sum = 0
            for a_out in range(alice_out):
                sig_a_sum += sigma[a_out, x_in]
            constraints.append(sig_a_sum == tau)

        # Enforce that:
        # \sum_{b \in \Gamma_B} \rho_b^y = \tau
        for y_in in range(bob_in):
            rho_b_sum = 0
            for b_out in range(bob_out):
                rho_b_sum += rho[b_out, y_in]
            constraints.append(rho_b_sum == tau)

        # Enforce that tau is a density operator.
        constraints.append(cvxpy.trace(tau) == 1)
        constraints.append(tau >> 0)

        problem = cvxpy.Problem(objective, constraints)
        ns_val = problem.solve()

        return ns_val

    def quantum_value_lower_bound(self, iters: int = 5, tol: float = 10e-6) -> float:
        r"""
        Calculate lower bound on the quantum value of an extended nonlocal game.

        Test

        :return: The quantum value of the extended nonlocal game.
        """
        # Get number of inputs and outputs for Bob's measurements.
        _, _, _, num_outputs_bob, _, num_inputs_bob = self.pred_mat.shape

        best_lower_bound = float("-inf")
        for _ in range(iters):
            # Generate a set of random POVMs for Bob. These measurements serve as a
            # rough starting point for the alternating projection algorithm.
            bob_povms = defaultdict(int)
            for y_ques in range(num_inputs_bob):
                random_mat = random_unitary(num_outputs_bob)
                for b_ans in range(num_outputs_bob):
                    random_mat_trans = random_mat[:, b_ans].conj().T.reshape(-1, 1)
                    bob_povms[y_ques, b_ans] = random_mat[:, b_ans] * random_mat_trans

            # Run the alternating projection algorithm between the two SDPs.
            it_diff = 1
            prev_win = -1
            best = float("-inf")
            while it_diff > tol:
                # Optimize over Alice's measurement operators while fixing Bob's.
                # If this is the first iteration, then the previously randomly
                # generated operators in the outer loop are Bob's. Otherwise, Bob's
                # operators come from running the next SDP.
                rho, lower_bound = self.__optimize_alice(bob_povms)
                bob_povms, lower_bound = self.__optimize_bob(rho)

                it_diff = lower_bound - prev_win
                prev_win = lower_bound

                # As the SDPs keep alternating, check if the winning probability
                # becomes any higher. If so, replace with new best.
                if best < lower_bound:
                    best = lower_bound

            if best_lower_bound < best:
                best_lower_bound = best

        return best_lower_bound

    def __optimize_alice(self, bob_povms) -> Tuple[Dict, float]:
        """Fix Bob's measurements and optimize over Alice's measurements."""
        # Get number of inputs and outputs.
        (
            dim,
            _,
            num_outputs_alice,
            num_outputs_bob,
            num_inputs_alice,
            num_inputs_bob,
        ) = self.pred_mat.shape

        # The cvxpy package does not support optimizing over 4-dimensional objects.
        # To overcome this, we use a dictionary to index between the questions and
        # answers, while the cvxpy variables held at this positions are
        # `dim`-by-`dim` cvxpy variables.
        rho = defaultdict(cvxpy.Variable)
        for x_ques in range(num_inputs_alice):
            for a_ans in range(num_outputs_alice):
                rho[x_ques, a_ans] = cvxpy.Variable(
                    (dim * num_outputs_bob, dim * num_outputs_bob), hermitian=True
                )

        tau = cvxpy.Variable((dim * num_outputs_bob, dim * num_outputs_bob), hermitian=True)
        win = 0
        for x_ques in range(num_inputs_alice):
            for y_ques in range(num_inputs_bob):
                for a_ans in range(num_outputs_alice):
                    for b_ans in range(num_outputs_bob):
                        if isinstance(bob_povms[y_ques, b_ans], np.ndarray):
                            win += self.prob_mat[x_ques, y_ques] * cvxpy.trace(
                                (
                                    np.kron(
                                        self.pred_mat[:, :, a_ans, b_ans, x_ques, y_ques],
                                        bob_povms[y_ques, b_ans],
                                    )
                                )
                                .conj()
                                .T
                                @ rho[x_ques, a_ans]
                            )
                        if isinstance(
                            bob_povms[y_ques, b_ans],
                            cvxpy.expressions.variable.Variable,
                        ):
                            win += self.prob_mat[x_ques, y_ques] * cvxpy.trace(
                                (
                                    np.kron(
                                        self.pred_mat[:, :, a_ans, b_ans, x_ques, y_ques],
                                        bob_povms[y_ques, b_ans].value,
                                    )
                                )
                                .conj()
                                .T
                                @ rho[x_ques, a_ans]
                            )
        objective = cvxpy.Maximize(cvxpy.real(win))
        constraints = list()

        # Sum over "a" for all "x" for Alice's measurements.
        for x_ques in range(num_inputs_alice):
            rho_sum_a = 0
            for a_ans in range(num_outputs_alice):
                rho_sum_a += rho[x_ques, a_ans]
                constraints.append(rho[x_ques, a_ans] >> 0)
            constraints.append(rho_sum_a == tau)

        constraints.append(cvxpy.trace(tau) == 1)
        constraints.append(tau >> 0)

        problem = cvxpy.Problem(objective, constraints)

        lower_bound = problem.solve()
        return rho, lower_bound

    def __optimize_bob(self, rho) -> Tuple[Dict, float]:
        """Fix Alice's measurements and optimize over Bob's measurements."""
        # Get number of inputs and outputs.
        (
            dim,
            _,
            num_outputs_alice,
            num_outputs_bob,
            num_inputs_alice,
            num_inputs_bob,
        ) = self.pred_mat.shape

        # The cvxpy package does not support optimizing over 4-dimensional objects.
        # To overcome this, we use a dictionary to index between the questions and
        # answers, while the cvxpy variables held at this positions are
        # `dim`-by-`dim` cvxpy variables.
        bob_povms = defaultdict(cvxpy.Variable)
        for y_ques in range(num_inputs_bob):
            for b_ans in range(num_outputs_bob):
                bob_povms[y_ques, b_ans] = cvxpy.Variable((dim, dim), hermitian=True)
        win = 0
        for x_ques in range(num_inputs_alice):
            for y_ques in range(num_inputs_bob):
                for a_ans in range(num_outputs_alice):
                    for b_ans in range(num_outputs_bob):
                        win += self.prob_mat[x_ques, y_ques] * cvxpy.trace(
                            (
                                cvxpy.kron(
                                    self.pred_mat[:, :, a_ans, b_ans, x_ques, y_ques],
                                    bob_povms[y_ques, b_ans],
                                )
                            )
                            @ rho[x_ques, a_ans].value
                        )
        objective = cvxpy.Maximize(cvxpy.real(win))

        constraints = list()

        # Sum over "b" for all "y" for Bob's measurements.
        for y_ques in range(num_inputs_bob):
            bob_sum_b = 0
            for b_ans in range(num_outputs_bob):
                bob_sum_b += bob_povms[y_ques, b_ans]
                constraints.append(bob_povms[y_ques, b_ans] >> 0)
            constraints.append(bob_sum_b == np.identity(num_outputs_bob))

        problem = cvxpy.Problem(objective, constraints)

        lower_bound = problem.solve()
        return bob_povms, lower_bound



"""Helper functions for checking validity of states and probability vectors."""
from typing import List

import numpy as np


def __is_states_valid(states: List[np.ndarray]) -> bool:
    """Check if states provided are valid."""
    # Assume that at least one state is provided.
    if states is None or states == []:
        raise ValueError("InvalidStates: There must be at least one state provided.")
    return True


def __is_probs_valid(probs: List[float]) -> bool:
    """Check if probabilities provided are valid."""
    if not np.isclose(sum(probs), 1):
        raise ValueError("InvalidProbabilities: Probabilities must sum to 1.")
    return True



"""Test is_mutually_unbiased_basis."""
import numpy as np

from toqito.state_props import is_mutually_unbiased_basis
from toqito.states import basis


def test_is_mutually_unbiased_basis_dim_2():
    """Return True for MUB of dimension 2."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    mub_1 = [e_0, e_1]
    mub_2 = [1 / np.sqrt(2) * (e_0 + e_1), 1 / np.sqrt(2) * (e_0 - e_1)]
    mub_3 = [1 / np.sqrt(2) * (e_0 + 1j * e_1), 1 / np.sqrt(2) * (e_0 - 1j * e_1)]
    mubs = [mub_1, mub_2, mub_3]
    np.testing.assert_equal(is_mutually_unbiased_basis(mubs), True)


def test_is_not_mub_dim_2():
    """Return False for non-MUB of dimension 2."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    mub_1 = [e_0, e_1]
    mub_2 = [1 / np.sqrt(2) * (e_0 + e_1), e_1]
    mub_3 = [1 / np.sqrt(2) * (e_0 + 1j * e_1), e_0]
    mubs = [mub_1, mub_2, mub_3]
    np.testing.assert_equal(is_mutually_unbiased_basis(mubs), False)


def test_is_mutually_unbiased_basis_invalid_input_len():
    """Tests for invalid input len."""
    with np.testing.assert_raises(ValueError):
        vec_list = [np.array([1, 0])]
        is_mutually_unbiased_basis(vec_list)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Maximally entangled state."""
import numpy as np

from scipy import sparse

from toqito.matrices import iden


def max_entangled(
    dim: int, is_sparse: bool = False, is_normalized: bool = True
) -> [np.ndarray, sparse.dia.dia_matrix]:
    r"""
    Produce a maximally entangled bipartite pure state [WikEnt]_.

    Produces a maximally entangled pure state as above that is sparse if :code:`is_sparse = True`
    and is full if :code:`is_sparse = False`. The pure state is normalized to have Euclidean norm 1
    if :code:`is_normalized = True`, and it is unnormalized (i.e. each entry in the vector is 0 or 1
    and the Euclidean norm of the vector is :code:`sqrt(dim)` if :code:`is_normalized = False`.

    Examples
    ==========

    We can generate the canonical :math:`2`-dimensional maximally entangled state

    .. math::
        u = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right)

    using :code:`toqito` as follows.

    >>> from toqito.states import max_entangled
    >>> max_entangled(2)
    [[0.70710678],
     [0.        ],
     [0.        ],
     [0.70710678]]

    By default, the state returned in normalized, however we can generate the unnormalized state

    .. math::
        v = |00\rangle + |11 \rangle

    using :code:`toqito` as follows.

    >>> from toqito.states import max_entangled
    >>> max_entangled(2, False, False)
    [[1.],
     [0.],
     [0.],
     [1.]]

    References
    ==========
    .. [WikEnt] Wikipedia: Quantum entanglement
        https://en.wikipedia.org/wiki/Quantum_entanglement

    :param dim: Dimension of the entangled state.
    :param is_sparse: `True` if vector is spare and `False` otherwise.
    :param is_normalized: `True` if vector is normalized and `False` otherwise.
    :return: The maximally entangled state of dimension :code:`dim`.
    """
    psi = np.reshape(iden(dim, is_sparse), (dim ** 2, 1))
    if is_normalized:
        psi = psi / np.sqrt(dim)
    return psi



"""Optimizations over quantum states."""
from toqito.state_opt.optimal_clone import optimal_clone
from toqito.state_opt.ppt_distinguishability import ppt_distinguishability
from toqito.state_opt.state_distinguishability import state_distinguishability
from toqito.state_opt.state_exclusion import state_exclusion
from toqito.state_opt.symmetric_extension_hierarchy import symmetric_extension_hierarchy



"""Swap operator."""
from typing import List, Union

import numpy as np

from toqito.matrices import iden
from toqito.perms import swap


def swap_operator(dim: Union[List[int], int], is_sparse: bool = False) -> np.ndarray:
    r"""
    Produce a unitary operator that swaps two subsystems.

    Provides the unitary operator that swaps two copies of :code:`dim`-dimensional space. If the two
    subsystems are not of the same dimension, :code:`dim` should be a 1-by-2 vector containing the
    dimension of the subsystems.

    Examples
    ==========

    The :math:`2`-dimensional swap operator is given by the following matrix

    .. math::
        X_2 =
        \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 0 & 1 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 0 & 1
        \end{pmatrix}

    Using :code:`toqito` we can obtain this matrix as follows.

    >>> from toqito.perms import swap_operator
    >>> swap_operator(2)
    [[1., 0., 0., 0.],
     [0., 0., 1., 0.],
     [0., 1., 0., 0.],
     [0., 0., 0., 1.]]

    The :math:`3`-dimensional operator may be obtained using :code:`toqito` as follows.

    >>> from toqito.perms import swap_operator
    >>> swap_operator(3)
    [[1., 0., 0., 0., 0., 0., 0., 0., 0.],
     [0., 0., 0., 1., 0., 0., 0., 0., 0.],
     [0., 0., 0., 0., 0., 0., 1., 0., 0.],
     [0., 1., 0., 0., 0., 0., 0., 0., 0.],
     [0., 0., 0., 0., 1., 0., 0., 0., 0.],
     [0., 0., 0., 0., 0., 0., 0., 1., 0.],
     [0., 0., 1., 0., 0., 0., 0., 0., 0.],
     [0., 0., 0., 0., 0., 1., 0., 0., 0.],
     [0., 0., 0., 0., 0., 0., 0., 0., 1.]]

    :param dim: The dimensions of the subsystems.
    :param is_sparse: Sparse if :code:`True` and non-sparse if :code:`False`.
    :return: The swap operator of dimension :code:`dim`.
    """
    # Allow the user to enter a single number for dimension.
    if isinstance(dim, int):
        dim = np.array([dim, dim])

    # Swap the rows of the identity appropriately.
    return swap(iden(int(np.prod(dim)), is_sparse), [1, 2], dim, True)



"""Test update_odometer."""
import numpy as np

from toqito.helper import update_odometer


def test_update_odometer_0_0():
    """Update odometer from [2, 2] to [0, 0]."""
    vec = np.array([2, 2])
    upper_lim = np.array([3, 2])
    res = update_odometer(vec, upper_lim)

    bool_mat = np.isclose([0, 0], res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_update_odometer_0_1():
    """Update odometer from [0, 0] to [0, 1]."""
    vec = np.array([0, 0])
    upper_lim = np.array([3, 2])
    res = update_odometer(vec, upper_lim)

    bool_mat = np.isclose([0, 1], res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_update_odometer_1_0():
    """Update odometer from [0, 1] to [1, 0]."""
    vec = np.array([0, 1])
    upper_lim = np.array([3, 2])
    res = update_odometer(vec, upper_lim)

    bool_mat = np.isclose([1, 0], res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_update_odometer_2_0():
    """Update odometer from [1, 1] to [2, 0]."""
    vec = np.array([1, 1])
    upper_lim = np.array([3, 2])
    res = update_odometer(vec, upper_lim)

    bool_mat = np.isclose([2, 0], res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_update_odometer_2_1():
    """Update odometer from [2, 0] to [2, 1]."""
    vec = np.array([2, 0])
    upper_lim = np.array([3, 2])
    res = update_odometer(vec, upper_lim)

    bool_mat = np.isclose([2, 1], res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_update_odometer_2_2():
    """Update odometer from [2, 1] to [0, 0]."""
    vec = np.array([2, 1])
    upper_lim = np.array([3, 2])
    res = update_odometer(vec, upper_lim)

    bool_mat = np.isclose([0, 0], res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_update_odometer_empty():
    """Return `None` if empty lists are provided."""
    vec = np.array([])
    upper_lim = np.array([])
    res = update_odometer(vec, upper_lim)

    bool_mat = np.isclose([], res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Checks whether operator is in the ball of separability centered at the maximally-mixed state."""
import numpy as np


def in_separable_ball(mat: np.ndarray) -> bool:
    r"""
    Check whether an operator is contained in ball of separability [GB02]_.

    Determines whether :code:`mat` is contained within the ball of separable operators centered
    at the identity matrix (i.e. the maximally-mixed state). The size of this ball was derived in
    [GB02]_.

    This function can be used as a method for separability testing of states in certain scenarios.

    This function is adapted from QETLAB.

    Examples
    ==========

    The only states acting on :math:`\complex^m \otimes \complex^n` in the separable ball that do
    not have full rank are those with exactly 1 zero eigenvalue, and the :math:`mn - 1` non-zero
    eigenvalues equal to each other.

    The following is an example of generating a random density matrix with eigenvalues
    :code:`[1, 1, 1, 0]/3`. This example yields a matrix that is contained within the separable
    ball.

    >>> from toqito.random import random_unitary
    >>> from toqito.state_props import in_separable_ball
    >>> import numpy as np
    >>>
    >>> U = random_unitary(4)
    >>> lam = np.array([1, 1, 1, 0]) / 3
    >>> rho = U @ np.diag(lam) @ U.conj().T
    >>> in_separable_ball(rho)
    True

    The following is an example of generating a random density matrix with eigenvalues
    :code:`[1.01, 1, 0.99, 0]/3`. This example yields a matrix that is not contained within the
    separable ball.

    >>> from toqito.random import random_unitary
    >>> from toqito.state_props import in_separable_ball
    >>> import numpy as np
    >>>
    >>> U = random_unitary(4)
    >>> lam = np.array([1.01, 1, 0,.99, 0]) / 3
    >>> rho = U @ np.diag(lam) @ U.conj().T
    >>> in_separable_ball(rho)
    False

    References
    ==========
    .. [GB02] Gurvits, Leonid, and Barnum, Howard.
        "Largest separable balls around the maximally mixed bipartite quantum state."
        Physical Review A 66.6 (2002): 062311.
        https://arxiv.org/pdf/quant-ph/0204159.pdf

    :param mat: A positive semidefinite matrix or a vector of the eigenvalues of a positive
                semidefinite matrix.
    :return: :code:`True` if the matrix :code:`mat` is contained within the separable ball, and
            :code:`False` otherwise.
    """
    mat_dims = mat.shape
    max_dim = max(mat_dims)

    # If the matrix is a vector, turn it into a matrix. We could instead turn every matrix into a
    # vector of eigenvalues, but that would make the computation take O(n^3) time instead of the
    # current method which is O(n^2).

    # Case: Vector of eigenvalues.
    if len(mat_dims) == 1 or min(mat_dims) == 1:
        mat = np.diag(mat)

    # If the matrix has trace equal to 0 or less, it cannot be in the separable ball.
    if np.trace(mat) < max_dim * np.finfo(float).eps:
        return False

    mat = mat / np.trace(mat)

    # The following check relies on the fact that we scaled the matrix so that trace(mat) = 1.
    # The following condition is then exactly the condition mentioned in [GB02]_.
    return np.linalg.norm(mat / np.linalg.norm(mat, "fro") ** 2 - np.eye(max_dim), "fro") <= 1



"""Test is_positive_definite."""
import numpy as np

from toqito.matrix_props import is_positive_definite


def test_is_is_positive_definite():
    """Check that positive definite matrix returns True."""
    mat = np.array([[2, -1, 0], [-1, 2, -1], [0, -1, 2]])
    np.testing.assert_equal(is_positive_definite(mat), True)


def test_is_not_positive_definite():
    """Check that non-positive definite matrix returns False."""
    mat = np.array([[-1, -1], [-1, -1]])
    np.testing.assert_equal(is_positive_definite(mat), False)


def test_is_positive_definite_not_hermitian():
    """Input must be a Hermitian matrix."""
    mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    np.testing.assert_equal(is_positive_definite(mat), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test basis."""
import numpy as np

from toqito.states import basis


def test_basis_ket_0():
    """Test for `|0>`."""
    expected_res = np.array([[1], [0]])
    res = basis(2, 0)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_basis_ket_1():
    """Test for `|1>`."""
    expected_res = np.array([[0], [1]])
    res = basis(2, 1)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_basis_ket_0000():
    """Test for `|0000>`."""
    expected_res = np.array([[1], [0], [0], [0]])
    res = basis(4, 0)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_basis_invalid_dim():
    """Tests for invalid dimension inputs."""
    with np.testing.assert_raises(ValueError):
        basis(4, 4)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""cvxpy expression as np.array."""
import numpy as np
from cvxpy.expressions.expression import Expression


def expr_as_np_array(cvx_expr: Expression) -> np.ndarray:
    """
    Convert cvxpy expression into a numpy array.

    :param cvx_expr: The cvxpy expression to be converted.
    :return: The numpy array of the cvxpy expression.
    """
    if cvx_expr.is_scalar():
        return np.array(cvx_expr)
    if len(cvx_expr.shape) == 1:
        return np.array(list(cvx_expr))
    # Then cvx_expr is a 2-D array.
    rows = []
    for i in range(cvx_expr.shape[0]):
        row = [cvx_expr[i, j] for j in range(cvx_expr.shape[1])]
        rows.append(row)
    arr = np.array(rows)
    return arr



"""np.array object as cvxpy expression object."""
import numpy as np
from cvxpy import bmat
from cvxpy.expressions.expression import Expression


def np_array_as_expr(np_arr: [np.ndarray]) -> Expression:
    """
    Convert numpy array into a cvxpy expression.

    :param np_arr: The numpy array to be converted.
    :return: The cvxpy expression of the numpy array.
    """
    as_list = np_arr.tolist()
    expr = bmat(as_list)
    return expr



"""Test permutation_operator."""
import numpy as np

from toqito.perms import permutation_operator


def test_permutation_operator_standard_swap():
    """Generates the standard swap operator on two qubits."""
    expected_res = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    res = permutation_operator(2, [2, 1])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_permutation_operator_standard_swap_list_dim():
    """Generates the standard swap operator on two qubits."""
    expected_res = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

    res = permutation_operator([2, 2], [2, 1])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_permutation_operator_sparse_option():
    """Sparse swap operator on two qutrits."""
    res = permutation_operator(3, [2, 1], False, True)

    np.testing.assert_equal(res[0][0], 1)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test hadamard."""
import numpy as np

from toqito.matrices import hadamard


def test_hadamard_0():
    """Test for Hadamard function when n = 0."""
    res = hadamard(0)
    expected_res = np.array([[1]])
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_hadamard_1():
    """Test for Hadamard function when n = 1."""
    res = hadamard(1)
    expected_res = 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]])
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_hadamard_2():
    """Test for Hadamard function when n = 2."""
    res = hadamard(2)
    expected_res = 1 / 2 * np.array([[1, 1, 1, 1], [1, -1, 1, -1], [1, 1, -1, -1], [1, -1, -1, 1]])
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_hadamard_3():
    """Test for Hadamard function when n = 3."""
    res = hadamard(3)
    expected_res = (
        1
        / (2 ** (3 / 2))
        * np.array(
            [
                [1, 1, 1, 1, 1, 1, 1, 1],
                [1, -1, 1, -1, 1, -1, 1, -1],
                [1, 1, -1, -1, 1, 1, -1, -1],
                [1, -1, -1, 1, 1, -1, -1, 1],
                [1, 1, 1, 1, -1, -1, -1, -1],
                [1, -1, 1, -1, -1, 1, -1, 1],
                [1, 1, -1, -1, -1, -1, 1, 1],
                [1, -1, -1, 1, -1, 1, 1, -1],
            ]
        )
    )
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_hadamard_negative():
    """Input must be non-negative."""
    with np.testing.assert_raises(ValueError):
        hadamard(-1)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Pauli matrices."""
from typing import List, Union
from scipy import sparse

import numpy as np

from toqito.matrix_ops import tensor


def pauli(
    ind: Union[int, str, List[int], List[str]], is_sparse: bool = False
) -> Union[np.ndarray, sparse.csr_matrix]:
    r"""
    Produce a Pauli operator [WikPauli]_.

    Provides the 2-by-2 Pauli matrix indicated by the value of `ind`. The
    variable `ind = 1` gives the Pauli-X operator, `ind = 2` gives the Pauli-Y
    operator, `ind =3` gives the Pauli-Z operator, and `ind = 0`gives the
    identity operator. Alternatively, `ind` can be set to "I", "X", "Y", or "Z"
    (case insensitive) to indicate the Pauli identity, X, Y, or Z operator.

    The 2-by-2 Pauli matrices are defined as the following matrices:

    .. math::

        \begin{equation}
            \begin{aligned}
                X = \begin{pmatrix}
                        0 & 1 \\
                        1 & 0
                    \end{pmatrix}, \quad
                Y = \begin{pmatrix}
                        0 & -i \\
                        i & 0
                    \end{pmatrix}, \quad
                Z = \begin{pmatrix}
                        1 & 0 \\
                        0 & -1
                    \end{pmatrix}, \quad
                I = \begin{pmatrix}
                        1 & 0 \\
                        0 & 1
                    \end{pmatrix}.
                \end{aligned}
            \end{equation}

    Examples
    ==========

    Example for identity Pauli matrix.

    >>> from toqito.matrices import pauli
    >>> pauli("I")
    [[1., 0.],
     [0., 1.]])

    Example for Pauli-X matrix.

    >>> from toqito.matrices import pauli
    >>> pauli("X")
    [[0, 1],
     [1, 0]])

    Example for Pauli-Y matrix.

    >>> from toqito.matrices import pauli
    >>> pauli("Y")
    [[ 0.+0.j, -0.-1.j],
     [ 0.+1.j,  0.+0.j]])

    Example for Pauli-Z matrix.

    >>> from toqito.matrices import pauli
    >>> pauli("Z")
    [[ 1,  0],
     [ 0, -1]])

    References
    ==========
    .. [WikPauli] Wikipedia: Pauli matrices
        https://en.wikipedia.org/wiki/Pauli_matrices

    :param ind: The index to indicate which Pauli operator to generate.
    :param is_sparse: Returns a sparse matrix if set to True and a non-sparse
                      matrix if set to False.
    """
    if isinstance(ind, (int, str)):
        if ind in ("x", "X", 1):
            pauli_mat = np.array([[0, 1], [1, 0]])
        elif ind in ("y", "Y", 2):
            pauli_mat = np.array([[0, -1j], [1j, 0]])
        elif ind in ("z", "Z", 3):
            pauli_mat = np.array([[1, 0], [0, -1]])
        else:
            pauli_mat = np.identity(2)

        if is_sparse:
            pauli_mat = sparse.csr_matrix(pauli_mat)

        return pauli_mat

    num_qubits = len(ind)
    pauli_mats = []
    for i in range(num_qubits - 1, -1, -1):
        pauli_mats.append(pauli(ind[i], is_sparse))
    return tensor(pauli_mats)



"""Test is_density."""
import numpy as np

from toqito.matrix_props import is_density
from toqito.random import random_density_matrix


def test_is_density_real_entries():
    """Test if random density matrix with real entries is density matrix."""
    mat = random_density_matrix(2, True)
    np.testing.assert_equal(is_density(mat), True)


def test_is_density_complex_entries():
    """Test if density matrix with complex entries is density matrix."""
    mat = random_density_matrix(4)
    np.testing.assert_equal(is_density(mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test cnot."""
import numpy as np

from toqito.matrices import cnot


def test_cnot():
    """Test standard CNOT gate."""
    res = cnot()
    expected_res = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Tests for ExtendedNonlocalGame class."""
import unittest
import numpy as np

from toqito.states import basis
from toqito.nonlocal_games.extended_nonlocal_game import ExtendedNonlocalGame


class TestExtendedNonlocalGame(unittest.TestCase):
    """Unit test for ExtendedNonlocalGame."""

    @staticmethod
    def bb84_extended_nonlocal_game():
        """Define the BB84 extended nonlocal game."""
        e_0, e_1 = basis(2, 0), basis(2, 1)
        e_p = (e_0 + e_1) / np.sqrt(2)
        e_m = (e_0 - e_1) / np.sqrt(2)

        dim = 2
        num_alice_out, num_bob_out = 2, 2
        num_alice_in, num_bob_in = 2, 2

        pred_mat = np.zeros([dim, dim, num_alice_out, num_bob_out, num_alice_in, num_bob_in])
        pred_mat[:, :, 0, 0, 0, 0] = e_0 * e_0.conj().T
        pred_mat[:, :, 0, 0, 1, 1] = e_p * e_p.conj().T
        pred_mat[:, :, 1, 1, 0, 0] = e_1 * e_1.conj().T
        pred_mat[:, :, 1, 1, 1, 1] = e_m * e_m.conj().T

        prob_mat = 1 / 2 * np.identity(2)

        return prob_mat, pred_mat

    @staticmethod
    def chsh_extended_nonlocal_game():
        """Define the CHSH extended nonlocal game."""
        dim = 2
        num_alice_out, num_bob_out = 2, 2
        num_alice_in, num_bob_in = 2, 2

        pred_mat = np.zeros([dim, dim, num_alice_out, num_bob_out, num_alice_in, num_bob_in])
        pred_mat[:, :, 0, 0, 0, 0] = np.array([[1, 0], [0, 0]])
        pred_mat[:, :, 0, 0, 0, 1] = np.array([[1, 0], [0, 0]])
        pred_mat[:, :, 0, 0, 1, 0] = np.array([[1, 0], [0, 0]])

        pred_mat[:, :, 1, 1, 0, 0] = np.array([[0, 0], [0, 1]])
        pred_mat[:, :, 1, 1, 0, 1] = np.array([[0, 0], [0, 1]])
        pred_mat[:, :, 1, 1, 1, 0] = np.array([[0, 0], [0, 1]])

        pred_mat[:, :, 0, 1, 1, 1] = 1 / 2 * np.array([[1, 1], [1, 1]])
        pred_mat[:, :, 1, 0, 1, 1] = 1 / 2 * np.array([[1, -1], [-1, 1]])

        prob_mat = np.array([[1 / 4, 1 / 4], [1 / 4, 1 / 4]])

        return prob_mat, pred_mat

    @staticmethod
    def moe_mub_4_in_3_out_game():
        """Define the monogamy-of-entanglement game defined by MUBs."""
        prob_mat = 1 / 4 * np.identity(4)

        dim = 3
        e_0, e_1, e_2 = basis(dim, 0), basis(dim, 1), basis(dim, 2)

        eta = np.exp((2 * np.pi * 1j) / dim)
        mub_0 = [e_0, e_1, e_2]
        mub_1 = [
            (e_0 + e_1 + e_2) / np.sqrt(3),
            (e_0 + eta ** 2 * e_1 + eta * e_2) / np.sqrt(3),
            (e_0 + eta * e_1 + eta ** 2 * e_2) / np.sqrt(3),
        ]
        mub_2 = [
            (e_0 + e_1 + eta * e_2) / np.sqrt(3),
            (e_0 + eta ** 2 * e_1 + eta ** 2 * e_2) / np.sqrt(3),
            (e_0 + eta * e_1 + e_2) / np.sqrt(3),
        ]
        mub_3 = [
            (e_0 + e_1 + eta ** 2 * e_2) / np.sqrt(3),
            (e_0 + eta ** 2 * e_1 + e_2) / np.sqrt(3),
            (e_0 + eta * e_1 + eta * e_2) / np.sqrt(3),
        ]

        # List of measurements defined from mutually unbiased basis.
        mubs = [mub_0, mub_1, mub_2, mub_3]

        num_in = 4
        num_out = 3
        pred_mat = np.zeros([dim, dim, num_out, num_out, num_in, num_in], dtype=complex)

        pred_mat[:, :, 0, 0, 0, 0] = mubs[0][0] * mubs[0][0].conj().T
        pred_mat[:, :, 1, 1, 0, 0] = mubs[0][1] * mubs[0][1].conj().T
        pred_mat[:, :, 2, 2, 0, 0] = mubs[0][2] * mubs[0][2].conj().T

        pred_mat[:, :, 0, 0, 1, 1] = mubs[1][0] * mubs[1][0].conj().T
        pred_mat[:, :, 1, 1, 1, 1] = mubs[1][1] * mubs[1][1].conj().T
        pred_mat[:, :, 2, 2, 1, 1] = mubs[1][2] * mubs[1][2].conj().T

        pred_mat[:, :, 0, 0, 2, 2] = mubs[2][0] * mubs[2][0].conj().T
        pred_mat[:, :, 1, 1, 2, 2] = mubs[2][1] * mubs[2][1].conj().T
        pred_mat[:, :, 2, 2, 2, 2] = mubs[2][2] * mubs[2][2].conj().T

        pred_mat[:, :, 0, 0, 3, 3] = mubs[3][0] * mubs[3][0].conj().T
        pred_mat[:, :, 1, 1, 3, 3] = mubs[3][1] * mubs[3][1].conj().T
        pred_mat[:, :, 2, 2, 3, 3] = mubs[3][2] * mubs[3][2].conj().T

        return prob_mat, pred_mat

    def test_bb84_unentangled_value(self):
        """Calculate the unentangled value of the BB84 game."""
        prob_mat, pred_mat = self.bb84_extended_nonlocal_game()
        bb84 = ExtendedNonlocalGame(prob_mat, pred_mat)
        res = bb84.unentangled_value()
        expected_res = np.cos(np.pi / 8) ** 2

        self.assertEqual(np.isclose(res, expected_res), True)

    def test_bb84_unentangled_value_rep_2(self):
        """Calculate the unentangled value for BB84 game for 2 repetitions."""
        prob_mat, pred_mat = self.bb84_extended_nonlocal_game()
        bb84_2 = ExtendedNonlocalGame(prob_mat, pred_mat, 2)
        res = bb84_2.unentangled_value()
        expected_res = np.cos(np.pi / 8) ** 4

        self.assertEqual(np.isclose(res, expected_res), True)

    def test_bb84_quantum_lower_bound(self):
        """Calculate the lower bound for the quantum value of theBB84 game."""
        prob_mat, pred_mat = self.bb84_extended_nonlocal_game()
        bb84 = ExtendedNonlocalGame(prob_mat, pred_mat)
        res = bb84.quantum_value_lower_bound()
        expected_res = np.cos(np.pi / 8) ** 2

        self.assertLessEqual(np.isclose(res, expected_res), True)

    def test_bb84_nonsignaling_value(self):
        """Calculate the non-signaling value of the BB84 game."""
        prob_mat, pred_mat = self.bb84_extended_nonlocal_game()
        bb84 = ExtendedNonlocalGame(prob_mat, pred_mat)
        res = bb84.nonsignaling_value()
        expected_res = np.cos(np.pi / 8) ** 2

        self.assertEqual(np.isclose(res, expected_res, rtol=1e-03), True)

    def test_bb84_nonsignaling_value_rep_2(self):
        """Calculate the non-signaling value of the BB84 game for 2 reps."""
        prob_mat, pred_mat = self.bb84_extended_nonlocal_game()
        bb84 = ExtendedNonlocalGame(prob_mat, pred_mat, 2)
        res = bb84.nonsignaling_value()
        expected_res = 0.73826

        self.assertEqual(np.isclose(res, expected_res, rtol=1e-03), True)

    def test_chsh_unentangled_value(self):
        """Calculate the unentangled value of the CHSH game."""
        prob_mat, pred_mat = self.chsh_extended_nonlocal_game()
        chsh = ExtendedNonlocalGame(prob_mat, pred_mat)
        res = chsh.unentangled_value()
        expected_res = 3 / 4

        self.assertEqual(np.isclose(res, expected_res), True)

    def test_moe_mub_4_in_3_out_unentangled_value(self):
        """
        Calculate the unentangled value of a monogamy-of-entanglement game.
        """
        prob_mat, pred_mat = self.moe_mub_4_in_3_out_game()
        moe = ExtendedNonlocalGame(prob_mat, pred_mat)
        res = moe.unentangled_value()
        expected_res = (3 + np.sqrt(5)) / 8

        self.assertEqual(np.isclose(res, expected_res), True)



"""Matrices."""
from toqito.matrices.clock import clock
from toqito.matrices.shift import shift
from toqito.matrices.cnot import cnot
from toqito.matrices.fourier import fourier
from toqito.matrices.gell_mann import gell_mann
from toqito.matrices.gen_gell_mann import gen_gell_mann
from toqito.matrices.gen_pauli import gen_pauli
from toqito.matrices.hadamard import hadamard
from toqito.matrices.iden import iden
from toqito.matrices.pauli import pauli



"""Pure to mixed operation."""
import numpy as np


def pure_to_mixed(phi: np.ndarray) -> np.ndarray:
    r"""
    Convert a state vector or density matrix to a density matrix.

    Examples
    ==========

    It is possible to convert a pure state vector to a mixed state vector using the :code:`toqito`
    package. Consider the following Bell state

    .. math::
        u = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right).

    The corresponding mixed state from :math:`u` is calculated as

    .. math::
        \rho = u u^* = \frac{1}{2} \begin{pmatrix}
                                        1 & 0 & 0 & 1 \\
                                        0 & 0 & 0 & 0 \\
                                        0 & 0 & 0 & 0 \\
                                        1 & 0 & 0 & 1
                                   \end{pmatrix}

    Using :code:`toqito`, we can obtain this matrix as follows.

    >>> from toqito.states import bell
    >>> from toqito.state_ops import pure_to_mixed
    >>> phi = bell(0)
    >>> pure_to_mixed(phi)
    [[0.5, 0. , 0. , 0.5],
     [0. , 0. , 0. , 0. ],
     [0. , 0. , 0. , 0. ],
     [0.5, 0. , 0. , 0.5]]

    We can also give matrix inputs to the function in :code:`toqito`.

    >>> from toqito.states import bell
    >>> from toqito.state_ops import pure_to_mixed
    >>> phi = bell(0) * bell(0).conj().T
    >>> pure_to_mixed(phi)
    [[0.5, 0. , 0. , 0.5],
     [0. , 0. , 0. , 0. ],
     [0. , 0. , 0. , 0. ],
     [0.5, 0. , 0. , 0.5]])

    :param phi: A density matrix or a pure state vector.
    :return: density matrix representation of :code:`phi`, regardless of whether :code:`phi` is
             itself already a density matrix or if if is a pure state vector.
    """
    # Compute the size of `phi`. If it's already a mixed state, leave it alone.
    # If it's a vector (pure state), make it into a density matrix.
    row_dim, col_dim = phi.shape[0], phi.shape[1]

    # It's a pure state vector.
    if min(row_dim, col_dim) == 1:
        return phi * phi.conj().T
    # It's a density matrix.
    if row_dim == col_dim:
        return phi
    # It's neither.
    raise ValueError("InvalidDim: `phi` must be either a vector or square " "matrix.")



"""Is matrix a symmetric matrix."""
import numpy as np

from toqito.matrix_props import is_square


def is_symmetric(mat: np.ndarray, rtol: float = 1e-05, atol: float = 1e-08) -> bool:
    r"""
    Determine if a matrix is symmetric [WikSym]_.

    The following 3x3 matrix is an example of a symmetric matrix:

    .. math::

        \begin{pmatrix}
            1 & 7 & 3 \\
            7 & 4 & -5 \\
            3 &-5 & 6
        \end{pmatrix}

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                1 & 7 & 3 \\
                7 & 4 & -5 \\
                3 & -5 & 6
            \end{pmatrix}

    our function indicates that this is indeed a symmetric matrix.

    >>> from toqito.matrix_props import is_symmetric
    >>> import numpy as np
    >>> A = np.array([[1, 7, 3], [7, 4, -5], [3, -5, 6]])
    >>> is_symmetric(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                1 & 2 \\
                4 & 5
            \end{pmatrix}

    is not symmetric.

    >>> from toqito.matrix_props import is_symmetric
    >>> import numpy as np
    >>> B = np.array([[1, 2], [3, 4]])
    >>> is_symmetric(B)
    False

    References
    ==========
    .. [WikSym] Wikipedia: Symmetric matrix
        https://en.wikipedia.org/wiki/Symmetric_matrix

    :param mat: The matrix to check.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: Returns :code:`True` if the matrix is symmetric and :code:`False` otherwise.
    """
    if not is_square(mat):
        return False
    return np.allclose(mat, mat.T, rtol=rtol, atol=atol)



"""GHZ state."""
from typing import List

import numpy as np

from scipy import sparse


def ghz(dim: int, num_qubits: int, coeff: List[int] = None) -> sparse:
    r"""
    Generate a (generalized) GHZ state [GHZ07]_.

    Returns a :code:`num_qubits`-partite GHZ state acting on :code:`dim` local dimensions, described
    in [GHZ07]_. For example, :code:`ghz(2, 3)` returns the standard 3-qubit GHZ state on qubits.
    The output of this function is sparse.

    For a system of :code:`num_qubits` qubits (i.e., :code:`dim = 2`), the GHZ state can be written
    as

    .. math::
        |GHZ \rangle = \frac{1}{\sqrt{n}} \left(|0\rangle^{\otimes n} +
        |1 \rangle^{\otimes n} \right)).

    Examples
    ==========

    When :code:`dim = 2`, and :code:`num_qubits = 3` this produces the standard GHZ state

    .. math::
        \frac{1}{\sqrt{2}} \left( |000 \rangle + |111 \rangle \right).

    Using :code:`toqito`, we can see that this yields the proper state.

    >>> from toqito.states import ghz
    >>> ghz(2, 3).toarray()
    [[0.70710678],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.70710678]]

    As this function covers the generalized GHZ state, we can consider higher dimensions. For
    instance here is the GHZ state in :math:`\mathbb{C}^{4^{\otimes 7}}` as

    .. math::
        \frac{1}{\sqrt{30}} \left(|0000000 \rangle + 2|1111111 \rangle +
        3|2222222 \rangle + 4|3333333\rangle \right).

    Using :code:`toqito`, we can see this generates the appropriate generalized GHZ state.

    >>> from toqito.states import ghz
    >>> ghz(4, 7, np.array([1, 2, 3, 4]) / np.sqrt(30)).toarray()
    [[0.18257419],
     [0.        ],
     [0.        ],
     ...,
     [0.        ],
     [0.        ],
     [0.73029674]])

    References
    ==========
    .. [GHZ07] Going beyond Bell's theorem.
        D. Greenberger and M. Horne and A. Zeilinger.
        E-print: [quant-ph] arXiv:0712.0921. 2007.

    :param dim: The local dimension.
    :param num_qubits: The number of parties (qubits/qudits)
    :param coeff: (default `[1, 1, ..., 1])/sqrt(dim)`:
                  a 1-by-`dim` vector of coefficients.
    :returns: Numpy vector array as GHZ state.
    """
    if coeff is None:
        coeff = np.ones(dim) / np.sqrt(dim)

    # Error checking:
    if dim < 1:
        raise ValueError("InvalidDim: `dim` must be at least 1.")
    if num_qubits < 1:
        raise ValueError("InvalidNumQubits: `num_qubits` must be at least 1.")
    if len(coeff) != dim:
        raise ValueError(
            "InvalidCoeff: The variable `coeff` must be a vector" " of length equal to `dim`."
        )

    # Construct the state (and do it in a way that is less memory-intensive
    # than naively tensoring things together.
    dim_sum = 1
    for i in range(1, num_qubits):
        dim_sum += dim ** i

    ret_ghz_state = sparse.lil_matrix((dim ** num_qubits, 1))
    for i in range(1, dim + 1):
        ret_ghz_state[(i - 1) * dim_sum] = coeff[i - 1]
    return ret_ghz_state



"""Tests for choi_to_kraus."""
import numpy as np

from toqito.channel_ops import choi_to_kraus


def test_choi_to_kraus():
    """Choi matrix of the swap operator."""

    choi_mat = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])
    kraus_ops = [
        np.array([[0, 1j / np.sqrt(2)], [-1j / np.sqrt(2), 0]]),
        np.array([[0, 1 / np.sqrt(2)], [1 / np.sqrt(2), 0]]),
        np.array([[1, 0], [0, 0]]),
        np.array([[0, 0], [0, 1]]),
    ]
    res_kraus_ops = choi_to_kraus(choi_mat)

    bool_mat = np.isclose(kraus_ops[0], res_kraus_ops[0])
    np.testing.assert_equal(np.all(bool_mat), True)

    bool_mat = np.isclose(kraus_ops[1], res_kraus_ops[1])
    np.testing.assert_equal(np.all(bool_mat), True)

    bool_mat = np.isclose(kraus_ops[2], res_kraus_ops[2])
    np.testing.assert_equal(np.all(bool_mat), True)

    bool_mat = np.isclose(kraus_ops[3], res_kraus_ops[3])
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Identity matrix."""
from scipy import sparse

import numpy as np


def iden(dim: int, is_sparse: bool = False) -> np.ndarray:
    r"""
    Calculate the `dim`-by-`dim` identity matrix [WIKID]_.

    Returns the `dim`-by-`dim` identity matrix. If `is_sparse = False` then
    the matrix will be full. If `is_sparse = True` then the matrix will be
    sparse.

    .. math::
        \mathbb{I} = \begin{pmatrix}
                        1 & 0 & 0 & \ldots & 0 \\
                        0 & 1 & 0 & \ldots & 0 \\
                        0 & 0 & 1 & \ldots & 0 \\
                        \vdots & \vdots & \vdots & \ddots & \vdots \\
                        0 & 0 & 0 & \ldots & 1
                   \end{pmatrix}

    Only use this function within other functions to easily get the correct
    identity matrix. If you always want either the full or the sparse
    identity matrix, just use numpy's built-in np.identity function.

    Examples
    ==========

    The identity matrix generated from :math:`d = 3` yields the following
    matrix:

    .. math::
        \mathbb{I}_3 = \begin{pmatrix}
            1 & 0 & 0 \\
            0 & 1 & 0 \\
            0 & 0 & 1
        \end{pmatrix}

    >>> from toqito.matrices import iden
    >>> iden(3)
    [[1., 0., 0.],
     [0., 1., 0.],
     [0., 0., 1.]])

    It is also possible to create sparse identity matrices. The sparse identity
    matrix generated from :math:`d = 10` yields the following matrix:

    >>> from toqito.matrices import iden
    >>> iden(10, True)
    <10x10 sparse matrix of type '<class 'numpy.float64'>' with 10 stored
    elements (1 diagonals) in DIAgonal format>

    References
    ==========
    .. [WIKID] Wikipedia: Identity matrix
        https://en.wikipedia.org/wiki/Identity_matrix

    :param dim: Integer representing dimension of identity matrix.
    :param is_sparse: Whether or not the matrix is sparse.
    :return: Sparse identity matrix of dimension `dim`.
    """
    if is_sparse:
        id_mat = sparse.eye(dim)
    else:
        id_mat = np.identity(dim)
    return id_mat



"""Generate random state vector."""
from typing import List, Union

import numpy as np

from toqito.states import max_entangled
from toqito.perms import swap


def random_state_vector(
    dim: Union[List[int], int], is_real: bool = False, k_param: int = 0
) -> np.ndarray:
    r"""Generate a random pure state vector.

    Examples
    ==========

    We may generate a random state vector. For instance, here is an example where we can generate a
    :math:`2`-dimensional random state vector.

    >>> from toqito.random import random_state_vector
    >>> vec = random_state_vector(2)
    >>> vec
    [[0.50993973+0.15292408j],
     [0.27787332+0.79960122j]]

    We can verify that this is in fact a valid state vector by computing the corresponding density
    matrix of the vector and checking if the density matrix is pure.

    >>> from toqito.state_props import is_pure
    >>> dm = vec.conj().T * vec
    >>> is_pure(dm)
    True

    :param dim: The number of rows (and columns) of the unitary matrix.
    :param is_real: Boolean denoting whether the returned matrix has real
                    entries or not. Default is :code:`False`.
    :param k_param: Default 0.
    :return: A :code:`dim`-by-:code:`dim` random unitary matrix.
    """
    # Schmidt rank plays a role.
    if 0 < k_param < np.min(dim):

        # Allow the user to enter a single number for dim.
        if isinstance(dim, int):
            dim = [dim, dim]

        # If you start with a separable state on a larger space and multiply
        # the extra `k_param` dimensions by a maximally entangled state, you
        # get a Schmidt rank `<= k_param` state.
        psi = max_entangled(k_param, True, False).toarray()

        a_param = np.random.rand(dim[0] * k_param, 1)
        b_param = np.random.rand(dim[1] * k_param, 1)

        if not is_real:
            a_param = a_param + 1j * np.random.rand(dim[0] * k_param, 1)
            b_param = b_param + 1j * np.random.rand(dim[1] * k_param, 1)

        mat_1 = np.kron(psi.conj().T, np.identity(int(np.prod(dim))))
        mat_2 = swap(
            np.kron(a_param, b_param),
            sys=[2, 3],
            dim=[k_param, dim[0], k_param, dim[1]],
        )

        ret_vec = mat_1 @ mat_2
        return np.divide(ret_vec, np.linalg.norm(ret_vec))

    # Schmidt rank is full, so ignore it.
    ret_vec = np.random.rand(dim, 1)
    if not is_real:
        ret_vec = ret_vec + 1j * np.random.rand(dim, 1)
    return np.divide(ret_vec, np.linalg.norm(ret_vec))



"""Is matrix a unitary matrix."""
import numpy as np

from toqito.matrix_props import is_square


def is_unitary(mat: np.ndarray, rtol: float = 1e-05, atol: float = 1e-08) -> bool:
    r"""
    Check if matrix is unitary [WikUnitary]_.

    A matrix is unitary if its inverse is equal to its conjugate transpose.

    Alternatively, a complex square matrix :math:`U` is unitary if its conjugate transpose
    :math:`U^*` is also its inverse, that is, if

    .. math::
        \begin{equation}
            U^* U = U U^* = \mathbb{I},
        \end{equation}

    where :math:`\mathbb{I}` is the identity matrix.

    Examples
    ==========

    Consider the following matrix

    .. math::
        X = \begin{pmatrix}
            0 & 1 \\
            1 & 0
            \end{pmatrix}

    our function indicates that this is indeed a unitary matrix.

    >>> from toqito.matrix_props import is_unitary
    >>> import numpy as np
    >>> A = np.array([[0, 1], [1, 0]])
    >>> is_unitary(A)
    True

    We may also use the `random_unitary` function from `toqito`, and can verify that a randomly
    generated matrix is unitary

    >>> from toqito.matrix_props import is_unitary
    >>> from toqito.random import random_unitary
    >>> mat = random_unitary(2)
    >>> is_unitary(mat)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
            1 & 0 \\
            1 & 1
            \end{pmatrix}

    is not unitary.

    >>> from toqito.matrix_props import is_unitary
    >>> import numpy as np
    >>> B = np.array([[1, 0], [1, 1]])
    >>> is_unitary(B)
    False

    References
    ==========
    .. [WikUnitary] Wikipedia: Unitary matrix.
        https://en.wikipedia.org/wiki/Unitary_matrix

    :param mat: Matrix to check.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: Return :code:`True` if matrix is unitary, and :code:`False` otherwise.
    """
    if not is_square(mat):
        return False

    uc_u_mat = mat.conj().T @ mat
    u_uc_mat = mat @ mat.conj().T
    id_mat = np.eye(len(mat))

    # If U^* * U = I U * U^*, the matrix "U" is unitary.
    return np.allclose(uc_u_mat, id_mat, rtol=rtol, atol=atol) and np.allclose(
        u_uc_mat, id_mat, rtol=rtol, atol=atol
    )
